import copy
import datetime
import json
import logging
import os
import threading

from django.utils.translation import gettext_lazy as _

from reface.core.errors import TerminalError
from reface.core.resh import Resh
from resh.utils import errors, const as resh_const
from resh.domains import system, bridge, gateway, module, config

logger = logging.getLogger('reface')

resh = Resh()


class Box:
    DEF_PORT = 55577

    FAKE = 0
    CONTROLLER = 1
    GATEWAY = 2

    UNKNOWN = 0
    CONNECTED = 1
    DISCONNECTED = 2
    ONLY_REFACE = 3
    ONLY_RESH = 4

    def __init__(self, id, name='', host='', port=DEF_PORT, role=GATEWAY):
        self._id = id
        self._modules = []
        self.name = name if name else f'Шкаф {id}'
        self.host = host
        self.port = port
        self.status = self.UNKNOWN
        self._role = self.CONTROLLER if id == 0 else role
        self._controller_type = None

    def __str__(self):
        return f'({self._id} {self.name}: {[m["address"] for m in self.modules]})'

    @property
    def id(self): return self._id

    @id.setter
    def id(self, _id): self._id = _id

    @property
    def modules(self): return self._modules

    @modules.setter
    def modules(self, _modules): self._modules = _modules

    @property
    def is_local(self): return self._id == 0

    @property
    def role(self): return self._role

    @role.setter
    def role(self, role): self._role = role

    @property
    def controller_type(self):
        if self._controller_type is None:
            self._controller_type = resh.send(system.get_device_type)
        return self._controller_type

    @classmethod
    def deserialize(cls, data: dict):
        return cls(**data)


class Controller(dict):
    @staticmethod
    def update_nginx():
        config.update_nginx().exec()

    def renew(self):
        requests = dict(IP=system.get_ip, device_type=system.get_device_type, dns=system.get_dns, datetime=system.get_date_time, mac=system.get_mac,
                        status=system.get_status, info=system.get_info)

        for key, cmd in requests.items():
            try:
                res = resh.send(cmd)
                if key == 'datetime':
                    d = datetime.datetime.fromisoformat(res['datetime'])
                    self['date'] = str(d.date())
                    self['time'] = str(d.time())
                    self['timezone'] = str(d.tzinfo)
                    self['timezones'] = [str(datetime.timezone(datetime.timedelta(hours=offset))) for offset in range(-12, 13)]
                elif isinstance(res, dict):
                    self.update(res)
                else:
                    self[key] = res
            except BaseException as ex:
                logging.error('"%s" exec error: %s' % (cmd.__name__, str(ex)))

    def set_date(self, date):
        if date != self['date']:
            _date = datetime.date.fromisoformat(date)
            resh.send(system.set_date, _date.day, _date.month, _date.year)
            self['date'] = date

    def set_time(self, time):
        if time != self['time']:
            resh.send(system.set_time, *time.split(':'))
            self['time'] = time

    def set_timezone(self, tz: str):
        if tz != self['timezone']:
            _tz = int(tz.split("UTC")[1].split(":")[0])
            _datetime = datetime.datetime.fromisoformat(self['date'] + 'T' + self['time'] + self['timezone'].replace('UTC', ''))
            _datetime = _datetime.astimezone(datetime.timezone(datetime.timedelta(hours=_tz)))
            resh.send(system.set_time_zone, _tz)
            self['date'] = str(_datetime.date())
            self['time'] = str(_datetime.time())
            self['timezone'] = tz

    def set_dns(self, dns):
        if dns != self['dns']:
            resh.send(system.set_dns, *dns)
            self['dns'] = dns

    def set_ip(self, ip, mask, gateway):
        if ip != self['IP'] or mask != self['Mask'] or gateway != self['Gateway']:
            resh.send(system.set_ip, ip, mask, gateway)
            self['IP'] = ip
            self['Mask'] = mask
            self['Gateway'] = gateway


class BoxesMeta(type):
    DEF_DUMP = '/tmp/reface_dump.json'
    _instance = None

    @property
    def instance(cls):
        if cls._instance is None:

            with open(os.environ.get('REFACE_CONFIG'), 'r') as file:
                conf = json.loads(file.read())

            dump = conf.get('dump', cls.DEF_DUMP)
            if dump == cls.DEF_DUMP:
                logger.warning('Dump file not set!')

            inst = Boxes(dump)
            inst.init()
            cls._instance = inst

        return cls._instance


class Boxes(dict, metaclass=BoxesMeta):
    def __init__(self, conf_file: str):
        super(Boxes, self).__init__()
        self._controller = None
        self.name = ''
        self.conf_file = conf_file
        self._timer = None
        self.configured = None
        self.is_sw_gateway = None

    def __getitem__(self, item) -> Box:
        return super(Boxes, self).__getitem__(item)

    def __start_timer(self):
        self.__stop_timer()
        self._timer = threading.Timer(1, self.__save)
        self._timer.start()

    def __stop_timer(self):
        if self._timer and isinstance(self._timer, threading.Timer):
            self._timer.cancel()
            self._timer = None

    def __save(self):
        logger.debug('save file')
        data = json.dumps(
            dict(is_sw_gateway=self.is_sw_gateway,
                 name=self.name,
                 boxes=[
                     dict(id=box.id, name=box.name, host=box.host, port=box.port)
                     for box in self.values() if box.status != Box.ONLY_RESH
                 ]),
            indent=4)
        try:
            with open(self.conf_file, 'w') as file:
                file.write(data)

            with open(self.conf_file + f'{os.path.extsep}double', 'w') as file:
                file.write(data)
        except BaseException as ex:
            logger.exception(ex)

    def __read(self):
        try:
            with open(self.conf_file, 'r') as file:
                res = file.read()
                if res:
                    return res
                else:
                    raise BaseException('Empty file... Try read double')
        except BaseException as ex:
            logger.error(ex)

            try:
                with open(f'{self.conf_file}{os.path.extsep}double', 'r') as file:
                    return file.read()
            except FileNotFoundError as ex:
                logger.warning(f'{self.conf_file} not exist. Create it...')
                return '{}'

    @property
    def has_controller(self): return isinstance(self._controller, Controller)

    def disconnect_sw_gateway(self):
        resh.send(bridge.stop)
        self.clear_type()

    def clear_type(self):
        os.remove(self.conf_file)
        os.remove(f'{self.conf_file}.double')
        self.init()

    def set_type(self, _type):
        if _type == 'gateway':
            try:
                resh.send(bridge.start)
            except BaseException as ex:
                logging.warning(ex)
            self.is_sw_gateway = True
            self.name = 'Софтовый шлюз'
        else:
            self.is_sw_gateway = False
            self.name = 'Новый объект'

        self.__save()
        self.init()

    @property
    def controller(self):
        if self._controller is None:
            self._controller = Controller()
            self._controller.renew()

        return self._controller

    def init(self):
        self.clear()

        conf = json.loads(self.__read())

        self.configured = bool(conf)

        if not self.configured:
            _bridge = resh.send(bridge.status)
            if _bridge and _bridge.get('status', False):
                self.configured = True
                self.name = 'Софтовый шлюз*'
                conf.setdefault('is_sw_gateway', True)

        if self.configured:
            self.is_sw_gateway = conf.get('is_sw_gateway', False)

            box = Box.deserialize({'id': 0})
            try:
                box.host = resh.send(system.get_ip)['IP']
            except BaseException as ex:
                logging.exception(ex)
            finally:
                self[0] = box

            if not self.is_sw_gateway:
                self.name = conf.get('name', 'Новый объект')
                try:
                    self.update_box(self[0])
                except (OSError, errors.CanOpenError) as ex:
                    logging.error(ex)

                for item in conf.get('boxes', []):
                    try:
                        box = Box.deserialize(item)
                    except BaseException as ex:
                        logger.exception(ex)
                    else:
                        try:
                            if box.is_local:
                                box.host = resh.send(system.get_ip)['IP']
                            self.update_box(box)
                        except TerminalError as ex:
                            logger.error(ex)
                        except BaseException as ex:
                            logger.error(ex)
                        self[box.id] = box

                try:
                    gateways = resh.send(gateway.status, box=self.local.id)
                    gateways = {int(id): item for id, item in gateways.items()}
                except BaseException as ex:
                    logger.error(ex)
                else:
                    for box in self.values():
                        if not box.is_local and box.id not in gateways:
                            box.status = Box.ONLY_REFACE

                    for index, _gateway in gateways.items():
                        if index not in self:
                            self[index] = Box(index)
                            try:
                                self[index].host = _gateway['address']
                                self[index].port = _gateway['address'].get('port', Box.DEF_PORT)
                                self[index].status = Box.ONLY_RESH
                                self.update_box(self[index])
                            except:
                                pass
                        elif _gateway and _gateway['interface']['ready'] and _gateway['address']['ready']:
                            self[index].status = Box.CONNECTED
                        else:
                            self[index].status = Box.DISCONNECTED

    def flush(self):
        self.__start_timer()

    def set_name(self, name: str):
        self.name = name
        self.flush()

    def update_box(self, box: Box):
        box.modules = resh.send(module.find, box=box.id)

    def readdress(self, box_id):
        module.readdress(resh_const.GATEWAY, box_id, resh_const.FORCED).exec()

    @property
    def local(self) -> Box: return self[0]

    def occupied(self):
        return [dict(index=box.id, host=box.host) for box in self]

    def set(self, box: Box):
        _gateway = resh.send(gateway.connect, box.id, box.host, box.port)
        if _gateway['interface']['ready'] and _gateway['address']['ready']:
            box.status = Box.CONNECTED
        else:
            box.status = Box.DISCONNECTED
        self.update_box(box)
        self[box.id] = box
        self.flush()

    def get(self, index: int) -> Box:
        return super(Boxes, self).get(index)

    def remove(self, box_id: int):
        if box_id == 0:
            for box in self.copy().values():
                if not box.is_local:
                    self.remove(box.id)
            self.clear_type()
        elif box_id in self:
            try:
                resh.send(gateway.disconnect, box_id)
            except BaseException as ex:
                raise ex
            else:
                self.__delitem__(box_id)
                self.flush()

    def set_box_name(self, box_id: int, name: str):
        if self[box_id].name != name:
            self[box_id].name = name
            self.flush()

    def get_indexes(self):
        occupied = [dict(index=0, host=self.local.host)] + [dict(index=box.id, host=box.host) for box in self.values()]
        first = 1
        while first in [occ['index'] for occ in occupied]:
            first += 1

        return occupied, first

    def module_info(self, box_id, module_id):
        data = dict(desc='Module not declared!')
        if box_id in self:
            for _module in self[box_id].modules:
                if _module['address'] == module_id:
                    data = copy.deepcopy(_module)
                    try:
                        data.update(resh.send(module.info, module_id, box=box_id))
                        data['series'] = data.pop('name', '')
                        data['name'] = data.pop('desc', '')
                        if _module['type'] == 1:
                            _module['channels'] = resh.send(module.dali_status, module_id, box=box_id)
                            data['model'] = 'RAPIDA-DA%s' % len(_module['channels'])
                            data['name'] = _('RPMDA-DA8')
                            data['channels'] = dict(
                                left=sorted(
                                    [ch for ch in _module['channels'] if ch['id'] % 2 > 0],
                                    key=lambda o: o['id'], reverse=True
                                ),
                                right=sorted(
                                    [ch for ch in _module['channels'] if ch['id'] % 2 == 0],
                                    key=lambda o: o['id'], reverse=True
                                )
                            )
                    except TerminalError as ex:
                        data['desc'] = str(ex)

        data['box'] = self[box_id]
        return data

    def get_modules(self, box_id):
        if box_id in self:
            return self[box_id].modules

        return []

    def set_module_address(self, box_id: int, module_id: int, address: int):
        resh.send(module.set_address, module_id, address, box=box_id)
        for _module in self[box_id].modules:
            if _module['address'] == module_id:
                _module['address'] = address
                break

        self[box_id].modules.sort(key=lambda m: m['address'])

    def update_module_firmware(self, box_id: int, module_id: int, firmware: str):
        resh.send(module.update_firmware, module_id, firmware, box=box_id)

    # def build_line(self, room, command_class, box, *args):
    #     def on_step(action, address):
    #         async_to_sync(layers.get_channel_layer().group_send)(
    #             room,
    #             {
    #                 'type': 'chat_message',
    #                 'message': {
    #                     action: address
    #                 }
    #             }
    #         )
    #
    #     _layer = layers.get_channel_layer()
    #     if room not in _layer.groups:
    #         command = command_class(*args)
    #         command.set_step_handler(on_step)
    #         conf.interface = box
    #
    #         threading.Thread(target=command.exec).start()
