import logging
import os
import re
import signal
import threading

from channels import layers
from django.contrib.auth.decorators import login_required
from django.shortcuts import render, redirect

from reface import const
from reface.core import errors
from reface.dashboard import forms
from reface.dashboard.models import Box, Boxes, Controller
from resh.protocols.dali_utils.external_commands.dali2 import events
from resh.utils import errors

FIRMWARE_DIR = '/tmp/rapida/firmwares/'
BUILDS_DIR = '/tmp/rapida/builds/'

logger = logging.getLogger('reface')


@login_required
def root(request):
    if not Boxes.instance.configured:
        return render(request, 'tools/configurator.html')
    if Boxes.instance.is_sw_gateway:
        return redirect('/rapida/0')
    return render(request, 'objects/system.html', context=dict(name=Boxes.instance.name,
                                                               boxes=sorted(Boxes.instance.values(), key=lambda box: box.id)))


@login_required
def upgrade(request):
    if request.method == 'POST':
        has_dependencies = False
        build = handle_uploaded_file(BUILDS_DIR, request.FILES['build'])
        for fs in request.FILES.getlist('dependencies', []):
            has_dependencies = True
            handle_uploaded_file(BUILDS_DIR, fs)

        cmd = 'sudo pip3 install %s --upgrade' % build
        if has_dependencies:
            cmd += ' --no-index --find-links=%s' % BUILDS_DIR

        def update_and_exit(command):
            os.system(command)
            os.kill(os.getpid(), signal.SIGTERM)

        thr = threading.Thread(target=update_and_exit, args=[cmd])
        thr.start()
        return redirect_page(request, '/rapida/', 20)

    return render(request, 'tools/version_upgrade.html')


def redirect_page(request, url, timeout):
    return render(request, 'tools/redirector.html', context=dict(url=url, timeout=timeout))


@login_required
def configure(request):
    Boxes.instance.set_type(request.GET['type'])
    return redirect('/rapida/')


@login_required
def reload(request):
    Boxes.instance.init()
    return redirect('/rapida/')


@login_required
def controller(request):
    if 'update' in request.GET:
        if Boxes.instance.has_controller:
            Boxes.instance.controller.renew()
    elif 'disconnect' in request.GET:
        Boxes.instance.disconnect_sw_gateway()
        return redirect('/rapida/')
    elif 'nginx' in request.GET:
        threading.Thread(target=Controller.update_nginx).start()
        return redirect_page(request, request.GET.get('next', request.path), 15)
    elif 'upgrade' in request.GET:
        if 'time' in request.GET:
            Boxes.instance.controller.set_time(request.GET['time'])
        if 'date' in request.GET:
            Boxes.instance.controller.set_date(request.GET['date'])
        if 'timezone' in request.GET:
            Boxes.instance.controller.set_timezone(request.GET['timezone'])
        if 'dns0' in request.GET or 'dns1' in request.GET:
            Boxes.instance.controller.set_dns(' '.join([request.GET.get('dns0', ''), request.GET.get('dns1', '')]).strip().split())
        if 'ip' in request.GET or 'mask' in request.GET or 'gateway' in request.GET:
            Boxes.instance.controller.set_ip(ip=request.GET.get('ip'), mask=request.GET.get('mask'), gateway=request.GET.get('gateway'))

    if request.GET:
        return redirect(request.path)

    return render(request, 'objects/controller.html', context=dict(name=Boxes.instance.name,
                                                                   is_sw_gateway=Boxes.instance.is_sw_gateway,
                                                                   controller=Boxes.instance.controller))


@login_required
def set_box_name(request, box_id):
    Boxes.instance.set_box_name(int(box_id), request.GET.get('name', 'unknown'))
    return redirect(request.GET.get('redirect', '/rapida/'))


@login_required
def set_name(request):
    Boxes.instance.set_name(request.GET.get('name', 'unknown'))
    return redirect(request.path[:request.path.index('set_name')])


@login_required
def add_box(request):
    occupied, index = Boxes.instance.get_indexes()
    action = request.GET.get('action', None)
    index = int(request.GET.get('index', index))
    host = re.sub(r'[\d+][:\d+]*$', '', request.get_host().split(':')[0])
    host = str(request.GET.get('host', host))
    port = int(request.GET.get('port', Box.DEF_PORT))
    name = request.GET.get('name', '')
    role = Box.FAKE

    box = Box(index, host=host, port=port, name=name, role=role)

    if action == 'add':
        box.role = Box.GATEWAY
        Boxes.instance.set(box)
        return redirect(f'/rapida/box/{index}')
    elif action == 'conf':
        box.role = Box.GATEWAY
        Boxes.instance[index].status = Box.CONNECTED
        Boxes.instance.flush()
        return redirect(f'/rapida/box/{index}')
    elif action == 'remove':
        Boxes.instance.remove(index)
        return redirect('/rapida/')
    elif action == 'commit':
        Boxes.instance.set_box_name(index, name)
        return redirect(f'/rapida/box/{index}')

    return render(request, 'objects/box.html', context=dict(name=Boxes.instance.name, box=box, occupied=occupied))


@login_required
def box(request, box_id):
    _box = Boxes.instance.get(int(box_id))
    err = {}
    if 'update' in request.GET:
        Boxes.instance.update_box(_box)
        return redirect(request.path)
    elif 'readdress_modules' in request.POST:
        try:
            Boxes.instance.readdress(box_id)
        except BaseException as ex:
            err['readdress_modules'] = str(ex)
        else:
            Boxes.instance.update_box(box=_box)
            return redirect(request.path)
    return render(request, 'objects/box.html', context=dict(
        name=Boxes.instance.name,
        box=_box,
        occupied=Boxes.instance.get_indexes(),
        err=err
    ))


@login_required
def channel(request, box_id, module_id, channel_id):
    try:
        _module = [m for m in Boxes.instance[int(box_id)].modules if m['address'] == int(module_id)][0]
    except IndexError:
        logging.error('module not found')
    else:
        return render(
            request, 'objects/channel.html',
            context=dict(
                name=Boxes.instance.name,
                box=int(box_id),
                box_name=Boxes.instance[int(box_id)].name,
                module=int(module_id),
                channel=int(channel_id),
                channels=(
                    ch['id']
                    for ch in _module.get('channels', [])
                    if ch.get('powered', False)
                )
            )
        )


@login_required
def line_builder(request, box_id, module_id, channel_id):
    form = forms.LineBuilder(*(request.POST,) if request.method == 'POST' else ())
    room = '%s/%s/%s' % (box_id, module_id, channel_id)
    if request.method == 'POST':
        if 'last_addressing' in request.POST:
            return redirect(request.path + 'process/')
        elif form.is_valid():
            try:
                command = form.command(request.POST.get('action'))
                try:
                    layers.get_channel_layer(const.LINE_BUILDER).build_line(
                        command, box_id, module_id, channel_id, form.mode, form.start_address
                    )
                except errors.AlreadyStarted:
                    pass
                return redirect(request.path + 'process/')
            except BaseException as ex:
                logging.error(ex)

    return render(request, 'tools/line_builder.html', context=dict(
        name=Boxes.instance.name, box=box_id, module=module_id, channel=channel_id, form=form, room=room,
        box_name=Boxes.instance[int(box_id)].name))


@login_required
def line_builder_process(request, box_id, module_id, channel_id):
    room = '%s/%s/%s' % (box_id, module_id, channel_id)
    return render(request, 'tools/line_building.html', context=dict(
        name=Boxes.instance.name, box=box_id, module=module_id, channel=channel_id, room=room,
        action_verbose=forms.LineBuilder.verbose_name(
            layers.get_channel_layer(const.LINE_BUILDER).action_name(box_id, module_id, channel_id)
        ),
        box_name=Boxes.instance[int(box_id)].name
    ))


@login_required
def ordering(request, box_id, module_id, channel_id, dali_type):
    room = '%s/%s/%s' % (box_id, module_id, channel_id)
    if dali_type in request.GET:
        addresses = [int(i[1:]) if i[1:].isdigit() else i[1:] for i in request.GET.get(dali_type).split(',') if len(i) > 0]
    else:
        addresses = range(0, 64)
    return render(
        request,
        'tools/ordering.html',
        context=dict(
            name=Boxes.instance.name, box=box_id, module=module_id, channel=channel_id, room=room,
            addr_range=range(0, 64),
            addresses=addresses,
            dali_type=dali_type,
            dali1_name=forms.dali1.Dali1Protocol.domain_name,
            dali2_name=forms.dali2.Dali2Protocol.domain_name,
            box_name=Boxes.instance[int(box_id)].name
        ))


@login_required
def probation(request, box_id, module_id, channel_id):
    room = '%s/%s/%s' % (box_id, module_id, channel_id)
    return render(
        request,
        'tools/probation.html',
        context=dict(
            name=Boxes.instance.name, box=box_id, module=module_id, channel=channel_id, room=room,
            addresses=range(0, 64),
            commands=(
                # Real command name, Displayed name
                (forms.dali1.gear.QueryActualLevel.__name__, forms.dali1.gear.QueryActualLevel.__name__),
                (forms.dali1.gear.QueryMaxLevel.__name__, forms.dali1.gear.QueryMaxLevel.__name__),
                (forms.dali1.gear.QueryMinLevel.__name__, forms.dali1.gear.QueryMinLevel.__name__),
                (forms.dali1.led.QueryDimmingCurve.__name__, forms.dali1.led.QueryDimmingCurve.__name__),
            ),
            box_name=Boxes.instance[int(box_id)].name
        ))


@login_required
def check_device(request, box_id, module_id, channel_id):
    room = '%s/%s/%s' % (box_id, module_id, channel_id)
    return render(
        request,
        'tools/check_device.html',
        context=dict(
            name=Boxes.instance.name, box=box_id, module=module_id, channel=channel_id, room=room,
            addresses=range(0, 64),
            commands=(
                # Real command name, Displayed name
                (forms.dali1.gear.QueryActualLevel.__name__, forms.dali1.gear.QueryActualLevel.__name__),
                (forms.dali1.gear.QueryMaxLevel.__name__, forms.dali1.gear.QueryMaxLevel.__name__),
                (forms.dali1.gear.QueryMinLevel.__name__, forms.dali1.gear.QueryMinLevel.__name__),
                (forms.dali1.led.QueryDimmingCurve.__name__, forms.dali1.led.QueryDimmingCurve.__name__),
            ),
            box_name=Boxes.instance[int(box_id)].name
        ))


@login_required
def devices(request, box_id, module_id, channel_id):
    form = forms.LineScanner(*(request.POST,) if request.method == 'POST' else ())

    return render(request, 'tools/devices.html', context=dict(
        form=form,
        room='%s/%s/%s' % (box_id, module_id, channel_id),
        name=Boxes.instance.name, box=box_id, module=module_id, channel=channel_id,
        addresses=range(0, 64),
        groups_0_15=range(0, 16),
        groups_16_31=range(16, 32),
        columns=form.columns(),
        columns_width=6,
        commands=form.commands(),
        commands_width=5,
        event_filters=events.types,
        box_name=Boxes.instance[int(box_id)].name
    ))


@login_required
def module(request, box_id, module_id):
    box_id = int(box_id)
    module_id = int(module_id)
    err = {}
    if 'upgrade' in request.POST:
        if 'address' in request.POST:
            address = int(request.POST.get('address'))
            if module_id != address:
                try:
                    Boxes.instance.set_module_address(box_id, module_id, address)
                except errors.DomainError as ex:
                    err['address'] = str(ex)
                else:
                    module_id = address
        if 'firmware' in request.FILES:
            firmware = handle_uploaded_file(FIRMWARE_DIR, request.FILES['firmware'])
            try:
                Boxes.instance.update_module_firmware(box_id, module_id, firmware)
            except errors.DomainError as ex:
                err['firmware'] = str(ex)
            else:
                return redirect(f'/rapida/{box_id}/{module_id}/')
            finally:
                os.remove(firmware)

        if not err:
            return redirect(f'/rapida/{box_id}/{module_id}/')
    return render(request, 'objects/module.html', context=dict(name=Boxes.instance.name,
                                                               box=box_id,
                                                               module=Boxes.instance.module_info(box_id, module_id),
                                                               modules=Boxes.instance.get_modules(box_id),
                                                               errors=err))


def handle_uploaded_file(dest, f):
    os.makedirs(dest, exist_ok=True)
    filename = os.path.join(dest, f.name)
    with open(filename, 'wb+') as dest:
        for chunk in f.chunks():
            dest.write(chunk)
    return filename
