import json
import logging

from asgiref.sync import async_to_sync
from channels.generic.websocket import WebsocketConsumer

from resh.protocols.payloads.rapida import ServicePacket
from resh.protocols.payloads.dali import DaliPayload
from reface.line.monitor.vendor import WSVendorCan


class DaliMonitorConsumer(WebsocketConsumer):
    _vendor = None
    dali_requests = None
    bus_name = 'dali'

    @classmethod
    def room_name(cls, box, module, channel):
        return f'{cls.bus_name}-{box}-{module}-{channel}'

    def __init__(self, *args, **kargs):
        super(DaliMonitorConsumer, self).__init__(*args, **kargs)
        self.box = None
        self.module = None
        self.channel = None

    @property
    def vendor(self) -> WSVendorCan:
        if DaliMonitorConsumer._vendor is None:
            DaliMonitorConsumer._vendor = WSVendorCan(self.handler())
        return DaliMonitorConsumer._vendor

    def connect(self):
        # Join room group
        async_to_sync(self.channel_layer.group_add)(
            self.room_name(**self.scope['url_route']['kwargs']),
            self.channel_name
        )

        self.accept()

        self.vendor.add_subscription(**self.scope['url_route']['kwargs'])

    def disconnect(self, close_code):
        # Leave room group
        async_to_sync(self.channel_layer.group_discard)(
            self.room_name(**self.scope['url_route']['kwargs']),
            self.channel_name
        )

        self.vendor.remove_subscription(**self.scope['url_route']['kwargs'])

    # Receive message from room group
    def chat_message(self, event):
        message = event['message']

        # Send message to WebSocket
        self.send(text_data=json.dumps({
            'message': message
        }))

    def handler(self):
        while True:
            try:
                direct, gw, module, channel, dump  = yield
                payload = dump.payload

                d = dict(
                    time=payload.time,
                    hex=payload.raw_hex,
                    mark=payload.mark,
                    desc=payload.name,
                    name=payload.cmd_name,
                    module=dump.module_address,
                    direct=direct,
                )

                if isinstance(payload, DaliPayload):
                    d.update(dict(
                        address=payload.address if payload.has_address else '',
                        instance=payload.instance if payload.has_instance else '',
                        channel=dump.channel
                    ))
                elif isinstance(payload, ServicePacket):
                    pass
                else:
                    continue

                async_to_sync(self.channel_layer.group_send)(
                    self.room_name(box=gw, module=module, channel=channel+1),
                    {
                        'type': 'chat_message',
                        'message': {**d}
                    }
                )
            except StopIteration:
                logging.debug("STOP vendering")
                break
            except BaseException as ex:
                logging.exception(ex)
