import logging
import threading
from threading import Thread

import can

from resh.protocols.can_utils.dali_packet import DaliDump
from resh.protocols.can_utils.rapida_packet import RapidaDump
from resh.protocols.can_utils.can_packet import CanDump


def coroutine(func):
    def inner(*args, **kwargs):
        g = func(*args, **kwargs)
        g.send(None)
        return g
    return inner


class WSVendorCan(threading.Thread):
    def __init__(self, handler):
        super(WSVendorCan, self).__init__(name=self.__class__.__name__)
        self.buses = {}
        self.delegator = self.delegator(handler)

    @coroutine
    def delegator(self, g):
        yield from g

    def add_subscription(self, box, module, *args, **kwargs):
        box = int(box)
        module = int(module)
        try:
            if box not in self.buses:
                bus = BusThread(box, self.delegator)
                self.buses[box] = bus
                bus.start()
            else:
                bus = self.buses[box]

            bus.add_module(module)

        except OSError as ex:
            logging.error(ex)

    def remove_subscription(self, box, module, *args, **kwargs):
        box = int(box)
        module = int(module)
        if box in self.buses:
            self.buses[box].remove_module(module)
            if len(self.buses[box].modules) == 0:
                self.buses.pop(box).stop()


class Listener:
    def __init__(self, group, handler):
        self.group = group
        self.handler = handler
        self.topics = {}

    def compare(self, _topic):
        for topic in self.topics:
            if topic == _topic:
                return True
            elif '#' in topic:
                if _topic.index(topic[:topic.index('#')]) != -1:
                    return True

        return False


class BusThread(Thread):
    CONTROLLER_ID = 31

    def __init__(self, gateway, delegator):
        super(BusThread, self).__init__(name='VendorBus_%s' % gateway)
        self.gateway = gateway
        self.delegator = delegator
        self._stopped = False
        self.modules = {}

        self._bus = can.Bus(
            channel=self.can_name,
            **dict(
                bustype='socketcan',
                can_filters=[]
            )
        )

    @property
    def can_name(self):
        return '%scan%s' % ('' if self.gateway == 0 else 'r', self.gateway)

    def add_module(self, module):
        self.modules.setdefault(module, 0)
        if self.modules[module] == 0:
            self.modules[module] += 1
            filters = self._bus.filters
            if filters is None:
                filters = []
            filters += [
                {
                    "can_id": (module << 5) + self.CONTROLLER_ID,
                    "can_mask": 1023,
                },
                {
                    "can_id": (self.CONTROLLER_ID << 5) + module,
                    "can_mask": 1023,
                }
            ]

            self._bus.set_filters(filters)

    def remove_module(self, module):
        if module not in self.modules:
            return

        self.modules[module] -= 1
        if self.modules[module] < 1:
            self._bus.set_filters(
                [f
                 for f in self._bus.filters
                 if f['can_id'] != (module << 5) + self.CONTROLLER_ID
                 and f['can_id'] != (self.CONTROLLER_ID << 5) + module]
            )

            self.modules.pop(module)

    def stop(self):
        logging.info('Stop listening %s' % self.can_name)
        self._stopped = True
        self._bus.stop_all_periodic_tasks()
        self._bus.shutdown()

    def run(self) -> None:
        while not self._stopped:
            try:
                logging.info('Start listening %s' % self.can_name)
                for msg in self._bus:
                    try:
                        dump = CanDump.can_dump(msg)

                        if ((msg.arbitration_id >> 5) & self.CONTROLLER_ID) == self.CONTROLLER_ID:
                            channels = (
                                i for i in range(0, 8)
                                if (msg.data[2] >> i) & 1 == 1
                            )
                            direction = 'Send'
                            module = msg.arbitration_id & self.CONTROLLER_ID
                        else:
                            direction = 'Dump'
                            module = (msg.arbitration_id >> 5) & self.CONTROLLER_ID
                            channels = [msg.data[1] & 0x07]

                        if isinstance(dump, RapidaDump):
                            channels = range(0, 8)
                        elif isinstance(dump, DaliDump):
                            if direction == 'Send':
                                channels = (i for i in range(0, 8) if (msg.data[2] >> i) & 1 == 1)
                            else:
                                channels = [msg.data[1] & 0x07]

                        for channel in channels:
                            try:
                                self.delegator.send((direction, self.gateway, module, channel, dump))
                            except BaseException as ex:
                                logging.exception(ex)
                                continue

                    except BaseException as ex:
                        logging.exception(ex)
            except KeyboardInterrupt:
                self.stop()
            except BaseException as ex:
                if not self._stopped:
                    logging.exception(ex)
