import logging
import textwrap
import sys
from time import sleep

import resh.domains
from resh.locales.loc_def import loc
from resh.utils import settings, errors, completion_setup, logger, const

logging.getLogger('can.interfaces.socketcan.socketcan').setLevel(logging.WARNING)

DOMAINS = {
    domain.domain_name: domain for domain in (
        resh.domains.BridgeDomain,
        resh.domains.Dali1Protocol,
        resh.domains.Dali2Protocol,
        resh.domains.GatewayDomain,
        resh.domains.Module,
        resh.domains.SystemDomain,
        resh.domains.ConfigDomain,
    )
}


def run(args=sys.argv):
    try:
        settings.conf.parse(args)
    except BaseException as ex:
        print(str(ex))
        sys.exit(-1)

    if len(args) == 1:
        if settings.conf.help:
            help_domains()
        elif settings.conf.version:
            logger.log.result(const.VER)
        elif settings.conf.file:
            sys.exit(file_mode(settings.conf.file))
        elif settings.conf.completion_mode is True:
            sys.exit(resh.utils.completion_setup.build())
        else:
            sys.exit(interactive_mode())
    else:
        domain_name = args[1]

        domain = DOMAINS.get(domain_name)

        if domain:
            method_name = args[2].lower() if len(args) > 2 else ''
            if method_name:
                method = None
                try:
                    method = domain.from_name(method_name)(*args[3:])
                    if settings.conf.help:
                        method.help()
                    else:
                        logger.log.result(
                            method.exec()
                        )
                except errors.DomainError as ex:
                    logger.log.exception(ex)
                    if not isinstance(ex, (errors.NonexistentLocale, errors.DeviceNotAnswered, errors.CanNotCommitted,
                                           errors.CanOpenError)):
                        if not logger.log.is_json and not isinstance(ex, errors.UnavailableMethod):
                            if ex.show_help:
                                method.help(desc=False)
                    if not settings.conf.interactive_mode:
                        sys.exit(ex.code)
                except errors.StopExec as ex:
                    print(f'BREAK: {ex}')
                except KeyboardInterrupt:
                    method.set_break(0, {})
                except BaseException as ex:
                    print(f'AHTUNG: {ex}')
            else:
                help_methods(domain, settings.conf.help)
                if not settings.conf.interactive_mode:
                    sys.exit(errors.UNAVAILABLE_METHOD)

        else:
            logger.log.exception(errors.UnavailableDomain(domain_name))
            if not logger.log.is_json:
                help_domains()
            if not settings.conf.interactive_mode:
                sys.exit(errors.UNAVAILABLE_DOMAIN)


def file_mode(path):
    with open(path, 'r') as file:
        cmds = file.read().strip().split('\n')

    for cmd in cmds:
        print(f'*************{cmd}*****************')
        run(['file_mode'] + cmd.split())
        print()


def interactive_mode():
    _domains = {domain: [cmd.__name__ for cmd in tracker.commands] for domain, tracker in resh.DOMAINS.items()}
    _domains['exit'] = []
    _domains['version'] = []
    from resh.utils.completer import Completer
    Completer.start(_domains)
    print(const.VER)
    print(loc.INTERACTIVE_HELLO)

    try:
        import select
        while sys.stdin in select.select([sys.stdin], [], [], 0)[0]:
            _input = sys.stdin.readline()
            if _input:
                on_input(_input, need_print=True)
            else:
                return 0
    except KeyboardInterrupt:
        return 0
    except:
        pass

    while True:
        try:
            on_input(input('%s>' % settings.conf.interface))
        except (EOFError, StopIteration):
            return 0
        except KeyboardInterrupt:
            print()
            return 0


def on_input(msg, need_print=False):
    msg = msg.strip()
    if msg:
        if need_print:
            print('%s>%s' % (settings.conf.interface, msg))

        if msg == 'exit':
            raise StopIteration()
        elif 'sleep ' in msg:
            tm = msg.split('sleep ')[1].strip()
            try:
                sleep(float(tm))
            except:
                print('use "sleep <float>" for delay')
        else:
            run([settings.INTERACTIVE] + msg.split(' '))


def help_methods(domain, need_desc):
    if not need_desc:
        print('Available methods:')

    if need_desc:
        max_length = 0
        groups = {}
        for m in domain.commands:
            if len(m.__name__) > max_length:
                max_length = len(m.__name__)
            if m.group not in groups:
                groups[m.group] = []

            groups[m.group].append(m)

        for g in groups:
            print(('*'*(max_length+5) + f' {g} ({len(groups[g])}) ').ljust(80, '*'))
            # print(f'{g} ({len(groups[g])}) '.ljust(80, '*'))
            # print(f'{g} ({len(groups[g])}) '.center(80, '*'))
            for m in groups[g]:
                desc = m.desc()
                desc = textwrap.fill(desc, width=80 - max_length - 5)
                desc = desc.replace('\n', '\n'.ljust(max_length+7, ' '))
                print(f'   {m.__name__} {": ".rjust(max_length+2-len(m.__name__), " ")}{desc}')
            print()
    else:
        [print(f'    {m.__name__}') for m in domain.commands]


def help_domains():
    print('Supported domains: \n    ' + '\n    '.join(DOMAINS.keys()))


if __name__ == '__main__':
    run(sys.argv)
