import signal

import dali.address
import dali.command
import dali.frame
from dali.device.general import StartQuiescentMode, StopQuiescentMode, Compare as Compare2
from dali.gear.general import Compare as Compare1

from resh.locales.loc_def import loc
from resh.protocols.can_utils.dali_packet import DaliSend
from resh.utils import groups, tools, const, errors, logger, settings


class Domain(groups.Group):
    domain_name = ''
    def_params = {}
    opt_params = {}
    var_params = {}

    def __init__(self, *args):
        if not settings.conf.interactive_mode:
            signal.signal(signal.SIGTERM, self.set_break)
            signal.signal(signal.SIGINT, self.set_break)
            if hasattr(signal, 'SIGQUIT'):
                signal.signal(signal.SIGQUIT, self.set_break)

        self._module = None
        self._channel = None
        self._params = None
        self.has_break = False
        self.error_code = None
        self._args = list(args)
        self.__step_handler = self._step_handler

    def exec(self):
        """Непосредсвенно выполнение запроса. Требует переопределения в классе реализации"""
        raise NotImplementedError()

    @classmethod
    def help(cls, desc=True, template=True, params=True, example=True):
        if desc:
            print(cls.desc())
        if template:
            template = cls.template()
            print(loc.CMD_USAGE)
            print(f'  {template}')
        if params:
            params = cls.params_info()
            if len(params) > 0:
                for p in params:
                    print(f'    {p}')
        if example and (len(cls.def_params) > 0 or len(cls.opt_params) > 0 or len(cls.var_params) > 0):
            print(loc.CMD_EXAMPLE)
            print(f'  {cls.example()}')

    @property
    def module(self) -> int:
        """Адрес модуля"""
        if self._module is None:
            if len(self._args) < 1:
                raise errors.ModuleAddressError()
            else:
                self._module = tools.parse_module(self._args[0])
        return self._module

    @property
    def channel(self):
        if self._channel is None:
            if len(self._args) < 2:
                raise errors.ChannelAddressError()
            else:
                self._channel = tools.parse_channel(self._args[1])
        return self._channel

    @classmethod
    def desc(cls) -> str:
        """Описание метода. Выводится в help`е"""

        return cls.loc(cls.domain_name, cls.__name__, const.DESC)

    @classmethod
    def params_info(cls):
        """Описание параметров, отображаемое в hеlp`е"""
        l = 0
        for param in cls.def_params:
            if len(param) > l:
                l = len(param)
        for param in cls.opt_params:
            if len(param) > l:
                l = len(param)
        for param in cls.var_params:
            if len(param) + 2 > l:
                l = len(param) + 2

        return [f'{p} {":".rjust(l-len(p) + 2, " ")} {cls.loc(const.DEF_PARAMS, p)}' for p in cls.def_params] + \
               [f'{p} {":".rjust(l-len(p) + 2, " ")} {cls.loc(cls.domain_name, cls.__name__, p)}' for p in cls.opt_params] + \
               [f'[{p}] {":".rjust(l-len(p)+0, " ")} {cls.loc(cls.domain_name, cls.__name__, p)}' for p in cls.var_params]

    @classmethod
    def template(cls):
        """Шаблон формирования, отображаемое в hеlp`е"""

        return ' '.join([cls.domain_name, cls.__name__] +
                        [f'<{p}>' for p in cls.def_params] +
                        [f'<{p}>' for p in cls.opt_params] +
                        [f'[{p}]' for p in cls.var_params])

    @classmethod
    def example(cls):
        """Пример команды, отображаемый в hеlp`е"""

        return ' '.join([cls.domain_name, cls.__name__] +
                        [f'{cls.def_params[p]}' for p in cls.def_params] +
                        [f'{cls.opt_params[p]}' for p in cls.opt_params] +
                        [f'{cls.var_params[p]}' for p in cls.var_params])

    @classmethod
    def loc(cls, *keys):
        """Получить локализайию из localization.loc по указанным ключам"""

        try:
            locale = loc.loc()
            for key in keys:
                locale = locale[key]
        except:
            return '>'.join(keys)
            # raise NonexistentLocale(*keys)
        return locale

    def set_break(self, sig, params):
        """Завершение работы при получнии сигнала"""
        self.has_break = True

    @property
    def params(self):
        """Получить параметры команды, передвнные из терминала"""

        if self._params is None:
            self._params = list(self._args[len(self.def_params):])

        return self._params

    @property
    def has_next(self):
        return len(self.params) > 0

    @property
    def next_param(self):
        """Получить первый неиспользованный параметр команды"""
        if len(self.params) == 0:
            raise errors.BadParameter()

        return self.__transform_param(self.params.pop(0))

    def indexed_param(self, index):
        """Получить параметр по индексу"""
        if len(self.params) <= index:
            raise errors.BadParameter()

        return self.__transform_param(self.params.pop(index))

    @property
    def next_param_recursive(self):
        """Получить первый неиспользованный параметр команды, без изъятия из стека"""
        if len(self.params) == 0:
            raise errors.BadParameter()

        return self.__transform_param(self.params[0])

    def _step_handler(self, *args, **kwargs):
        pass

    def set_step_handler(self, handler):
        self.__step_handler = handler

    def on_step(self, *args, **kwargs):
        if callable(self.__step_handler):
            try:
                self.__step_handler(*args, **kwargs)
            except BaseException as ex:
                print(str(ex))

    @staticmethod
    def __transform_param(val):
        """Приведение параметра к int, если это возможно"""
        if isinstance(val, int):
            return val
        elif '0x' in val.lower():
            return int(val, 16)
        elif val.isdigit():
            return int(val)
        elif val.upper() == const.MASK:
            return 0xff
        return val


class DaliProtocol(Domain):
    domain_name = ''
    _bytes_count = 0

    def __init__(self, *args):
        super(DaliProtocol, self).__init__(*args)

        self._address = None
        self._address_raw = None

    def start_quiescent_mode(self, address=None):
        if address is None:
            address = dali.address.BroadcastUnaddressed()
        self.send(StartQuiescentMode(address))

    def stop_quiescent_mode(self, address=None):
        if address is None:
            address = dali.address.BroadcastUnaddressed()

        self.send(StopQuiescentMode(address))

    @property
    def address(self):
        if not self._address:
            if len(self._args) < 3:
                raise errors.DeviceAddressError()
            else:
                self._address = tools.parse_address(self._args[2])
        return self._address

    def raw_command(self, need_answer=False):
        try:
            frame = bytes([self.next_param for _ in range(0, self._bytes_count)])
        except:
            errors.BadParameter()
        else:
            send_twice = len(self.params) > 0 and str(self.next_param).upper() == const.SEND_TWICE.upper()
            return self.__send(frame, send_twice=send_twice, need_answer=need_answer)

    def send(self, cmd):
        """Простое выполнение команды переданной как экземпляр класса"""

        logger.log.info(f'-> {str(cmd)}')
        try:
            result = self.__send(cmd.frame.pack, send_twice=cmd.is_config, need_answer=cmd.is_query,
                                 listen_service=isinstance(cmd, (Compare1, Compare2)))
        except errors.DeviceNotAnswered as ex:
            if cmd.response == dali.command.YesNoResponse:
                result = None
            else:
                raise ex
        if cmd.is_query:
            if result:
                result_str = " ".join([hex(b)[2:].upper().rjust(2, "0") for b in result])
                logger.log.info(f'    <- {result_str}')
                response = cmd.response(dali.frame.BackwardFrame(result))
            else:
                logger.log.info(f'    <- <empty>')
                response = cmd.response(None)

            return response

    def __send(self, end_frame, send_twice=False, need_answer=False, listen_service=False):
        """Выполнение команды и необходимых подготовок с самостоятельным изъятием параметров их стека"""

        msg = DaliSend(1, self.module, self.channel, end_frame)
        msg.send_twice = send_twice
        return msg.send(need_answer=need_answer, listen_service=listen_service)
