import json
import shutil
import os

import nginx

from resh.domains._domain import Domain
from resh.utils import groups, compat
from resh.utils.const import CONFIG

CONFIG_SV = """[fcgi-program:rapida_reface]
command=daphne -u /run/rapida_reface.sock --fd 0 --access-log - --proxy-headers reface.asgi:application
socket=tcp://localhost:55581
;socket_owner=chrism
;socket_mode=0700
;process_name=%(program_name)s_%(process_num)02d
;numprocs=5
directory=/var/rapida
;umask=022
;priority=999
autostart=true
autorestart=true
startsecs=3
startretries=3
;exitcodes=0
stopsignal=QUIT
;stopasgroup=false
;killasgroup=false
stopwaitsecs=20
user=root
redirect_stderr=false
stdout_logfile=/var/log/rapida/reface_stdout.log
stdout_logfile_maxbytes=3MB
stdout_logfile_backups=10
;stdout_capture_maxbytes=1MB
;stdout_events_enabled=false
stderr_logfile=/var/log/rapida/reface_stderr.log
stderr_logfile_maxbytes=1MB
stderr_logfile_backups=10
;stderr_capture_maxbytes=1MB
;stderr_events_enabled=false
environment=
	REFACE_CONFIG=/etc/rapida/reface/config.json,
	DJANGO_SETTINGS_MODULE=reface.settings
;serverurl=AUTO
"""


CONFIG_SV_PATH = '/etc/supervisor/conf.d/rapida_reface.conf'
CONFIG_RF_PATH = '/etc/rapida/reface/config.json'
CONFIG_RF = {
    "http_root": "rapida",
    "dump": "/etc/rapida/reface/dump.json"
}


@compat.add_metaclass(compat.DomainTracker)
class ConfigDomain(Domain, groups.Config):
    domain_name = CONFIG

    @staticmethod
    def reset(clear_data=False):
        if clear_data:
            shutil.rmtree('/etc/rapida/', ignore_errors=True)
            shutil.rmtree('/tmp/rapida/', ignore_errors=True)

            os.makedirs(os.path.dirname(CONFIG_RF_PATH), exist_ok=True)

            with open(CONFIG_RF_PATH, 'w') as fs:
                json.dump(CONFIG_RF, fs, indent=4)

        os.makedirs(os.path.dirname(CONFIG_SV_PATH), exist_ok=True)

        with open(CONFIG_SV_PATH, 'w') as fs:
            fs.write(CONFIG_SV)

        os.system('sudo supervisorctl reread')
        os.system('sudo supervisorctl update')
        os.system('sudo supervisorctl restart rapida_reface')


""" https://kiwi.throne.systems/index.php/Tron/Strata/Hardware/Rapida/Resh/Command/Config#Config """


class reset(ConfigDomain):
    def exec(self):
        if input('Are you sure? (y/N)> ').lower() not in ('y', 'yes', 'da'):
            print('Break')
            return

        self.reset(clear_data=True)


class update(ConfigDomain):
    def exec(self): self.reset(clear_data=False)


class update_bash(ConfigDomain):
    """
    Генерирует файл автозаполнения для bash
    """
    CONF_PATH = '/etc/bash_completion.d/resh'

    def exec(self):
        from resh.utils import completion_setup
        completion_setup.build(self.CONF_PATH)


class update_nginx(ConfigDomain):
    """
    Проверяет наличие файла конфигурации для nginx
    В случае отсутствия - генерирует дефолтный
    Добавляет отсутствующие атрибуты для spread и reface
    """
    CONF_PATH = '/etc/nginx/sites-available/rapida'

    SETT_DEF = {
        '/': {
            'root': '/var/rapida/webroot',
            'index': 'index.html',
            'try_files': '$uri $uri/ =404',
        },
        '/red/': {
            'include': '/etc/nginx/websocket_params',
            'proxy_pass': 'http://127.0.0.1:1880',
        },
        '/rapida/': {
            'include': '/etc/nginx/asgi_params',
            'proxy_pass': 'http://127.0.0.1:55581',
        },
        '/rapida/static/': {
            'alias': '/usr/local/rapida/reface/static/',
        },
        '/spread/': {
            'include': '/etc/nginx/asgi_params',
            'proxy_pass': 'http://127.0.0.1:55582',
        },
        '/spread/static/': {
            'alias': '/usr/local/spread/dashboard/static/',
        },
    }

    SETT_DIN = {
        '/rapida/': {
            'client_max_body_size': '32m',
        },
        '/spread/': {
            'client_max_body_size': '32m',
        },
    }

    def exec(self):
        has_str = ''
        try:
            conf = nginx.loadf(self.CONF_PATH).server
            has_str = conf.as_strings
        except (FileExistsError, FileNotFoundError):
            conf = nginx.Server()
            conf.add(
                nginx.Comment('Generated automatically by resh\n'),
                nginx.Key('listen', '80 default_server'),
                nginx.Key('server_name', '_'),
                nginx.Key('charset', 'utf-8'),
            )

        for loc_name, loc_data in self.SETT_DEF.items():
            try:
                conf.filter(nginx.Location.__name__, name=loc_name)[-1]
            except IndexError:
                conf.add(nginx.Location(loc_name, *(nginx.Key(k, v) for k, v in loc_data.items())))

        for loc_name, loc_data in self.SETT_DIN.items():
            try:
                location = conf.filter(nginx.Location.__name__, name=loc_name)[-1]
            except IndexError:
                location = nginx.Location(loc_name)
                conf.add(location)

            for k, v in loc_data.items():
                try:
                    location.filter(name=k)[-1].value = v
                except IndexError:
                    location.add(nginx.Key(k, v))

        if has_str != conf.as_strings:
            nginx.dumpf(conf, self.CONF_PATH)
            os.system('sudo service nginx restart')


class migrate(ConfigDomain):
    def exec(self):
        if 'migrate' not in self._args:
            import reface.manage
            from django.core.management.commands.migrate import Command
            Command().run_from_argv([reface.manage.__file__, 'migrate', '--database', 'auth_db'])
