from __future__ import annotations

import sys
import threading
from time import sleep, time

import dali.address
import dali.command
import dali.frame
from dali.gear import general as gear, led

from resh.domains._domain import DaliProtocol
from resh.locales.loc_def import loc
from resh.protocols.dali_utils.bus1 import Bus1
from resh.protocols.dali_utils.external_commands import dali as ext_dali
from resh.protocols.dali_utils.external_commands.dali import colour_control as cc
from resh.protocols.dali_utils.external_commands.dali import switching_function as sf
from resh.utils import groups, tools, errors, logger, const_values as V, compat, settings
from resh.utils.const import *


@compat.add_metaclass(compat.DomainTracker)
class Dali1Protocol(DaliProtocol):
    def exec(self): raise NotImplementedError()

    def_params = {MODULE: 1, CHANNEL: 2, DEVICE_ADDR: 'D3'}
    domain_name = DALI
    _bytes_count = 2

    def __init__(self, *args):
        super(Dali1Protocol, self).__init__(*args)
        self._device_types = None

    @property
    def device_types(self):
        if self._device_types is None:
            _types = {}
            try:
                _type = self.send(gear.QueryDeviceType(self.address))
                if _type.value.as_integer == 0xff:
                    while _type is not None:
                        try:
                            _type = None
                            _type = self.send(ext_dali.QueryNextDeviceType(self.address))
                        except errors.DeviceNotAnswered:
                            _types[0xff] = 'Unreadable type'
                            break
                        else:
                            if _type.value.as_integer == 0xfe:
                                break

                            _types[_type.value.as_integer] = str(_type)
                else:
                    _types = {_type.value.as_integer: str(_type)}
            except (errors.CanError, errors.CanCollision):
                pass

            if len(_types) == 0:
                _types[0xff] = 'Unreadable type'

            self._device_types = _types

        return self._device_types

    def enable_type(self, for_cmd):
        self.send(gear.EnableDeviceType(for_cmd._devicetype))

    def send(self, cmd: gear.command.Command):
        if cmd._devicetype > 0:
            self.enable_type(cmd)

        return super(Dali1Protocol, self).send(cmd)

    def _send(self, cmd_class):
        """Выполнение команды и необходимых подготовок с самостоятельным изъятием параметров их стека"""

        dtr_cmds = []
        try:
            if cmd_class._hasparam:
                _cmd = cmd_class(self.address, self.next_param)
            else:
                _cmd = cmd_class(self.address)

            if cmd_class._uses_dtr0:
                dtr_cmds.append(gear.DTR0(self.next_param))
            if cmd_class._uses_dtr1:
                dtr_cmds.append(gear.DTR1(self.next_param))
            if cmd_class._uses_dtr2:
                dtr_cmds.append(gear.DTR2(self.next_param))
        except:
            raise errors.BadParameter()

        if dtr_cmds:
            for dtr_cmd in dtr_cmds:
                self.send(dtr_cmd)

        return self.send(_cmd)

    def _read_dtr(self, index: int):
        if index == 0:
            return self._send(ext_dali.QueryContentDTR0)
        elif index == 1:
            return self._send(ext_dali.QueryContentDTR1)
        elif index == 2:
            return self._send(ext_dali.QueryContentDTR2)
        else:
            raise BaseException('can read only DTR0, DTR1 and DTR2!')

    def write_memory_bank(self, bank_id=None, bank_addr=None, value=None):
        try:
            if bank_id is None:
                bank_id = self.next_param
            if bank_addr is None:
                bank_addr = self.next_param
            if value is None:
                value = self.next_param
        except:
            raise errors.BadParameter()

        if isinstance(self.address, dali.address.Short):
            cmd = gear.WriteMemoryLocation
        else:
            cmd = gear.WriteMemoryLocationNoReply

        self.send(gear.EnableWriteMemory(self.address))
        self.send(gear.DTR1(bank_id))

        self.send(gear.DTR0(0x02))
        self.send(cmd(0x55))

        self.send(gear.DTR0(bank_addr))
        resp = self.send(cmd(value))

        self.send(gear.DTR0(0x02))
        self.send(cmd(0xFF))

        if cmd == gear.WriteMemoryLocation:
            return resp.value.pack

    def read_memory_bank(self, bank_id=None, bank_addr=None):
        try:
            if bank_id is None:
                bank_id = self.next_param
            dtr1 = gear.DTR1(bank_id)
            if bank_addr is None:
                bank_addr = self.next_param
            dtr0 = gear.DTR0(bank_addr)
        except:
            raise errors.BadParameter()

        self.send(dtr1)
        self.send(dtr0)

        resp = self.send(gear.ReadMemoryLocation(self.address))
        return resp.value.pack

    def query_device_types(self): return self.device_types

    def set_short_address(self, new_address, need_warning=True, raise_if_used=True):
        try:
            new_address = tools.parse_address(new_address).address
        except:
            raise errors.BadParameter()

        try:
            self.send(gear.QueryDeviceType(new_address))
        except errors.DeviceNotAnswered:
            pass
        else:
            answer = None
            if need_warning:
                answer = logger.log.question(f'Address {new_address} will be duplicated. Continue? (y/N)', timeout=5)

            if raise_if_used and answer not in ['YES', 'Y', 'yes', 'y', '1', 'da']:
                raise errors.AddressAlreadyInUse(new_address)

        new_address = (new_address << 1) | 1

        self.send(gear.DTR0(new_address))
        self.send(gear.SetShortAddress(self.address))


"""https://kiwi.throne.systems/index.php/Tron/Strata/Hardware/Rapida/Resh/Command/Dali#Base"""


class raw_command(Dali1Protocol, groups.Base):
    def_params = {MODULE: 1, CHANNEL: 2}
    opt_params = {BYTE_ADDR: '0x00', BYTE_OPCODE: '0xfe'}
    var_params = {SEND_TWICE: SEND_TWICE}
    def exec(self): self.raw_command(False)


class raw_command_dtr0(Dali1Protocol, groups.Base):
    def_params = {MODULE: 1, CHANNEL: 2}
    opt_params = {BYTE_ADDR: '0x00', BYTE_OPCODE: '0xFE', DTR0_VALUE: '0xFF'}
    var_params = {SEND_TWICE: SEND_TWICE}

    def exec(self):
        self.send(gear.DTR0(self.indexed_param(2)))
        self.raw_command()


class enable_device_type(Dali1Protocol, groups.Base):
    def_params = {MODULE: 1, CHANNEL: 2, DEVICE_ADDR: 'D12'}
    opt_params = {DEVICE_TYPE: '8'}
    def exec(self): self.send(gear.EnableDeviceType(self.next_param))


class raw_query(Dali1Protocol, groups.Base):
    def_params = {MODULE: 1, CHANNEL: 2}
    opt_params = {BYTE_ADDR: '0x01', BYTE_OPCODE: '0xA0'}
    def exec(self): return self.raw_command(True)


class raw_read_memory(Dali1Protocol, groups.Base):
    opt_params = {MEMORY_BANK: 0, MEMORY_LOCATION: 8}
    def exec(self): return self.read_memory_bank()


class raw_write_memory(Dali1Protocol, groups.Base):
    opt_params = {MEMORY_BANK: 0, MEMORY_LOCATION: 8, MEMORY_VALUE: "FF"}
    def exec(self): self.write_memory_bank()


class read_dtr0(Dali1Protocol, groups.Base):
    def exec(self): return self._read_dtr(0).value.as_integer


class read_dtr1(Dali1Protocol, groups.Base):
    def exec(self): return self._read_dtr(1).value.as_integer


class read_dtr2(Dali1Protocol, groups.Base):
    def exec(self): return self._read_dtr(2).value.as_integer


""" https://kiwi.throne.systems/index.php/Tron/Strata/Hardware/Rapida/Resh/Command/Dali#Standard """


class set_actual_level(Dali1Protocol, groups.Standard):
    opt_params = {LEVEL: 254}
    def exec(self): self.send(gear.DAPC(self.address, self.next_param))


class off(Dali1Protocol, groups.Standard):
    def exec(self): self._send(gear.Off)


class up(Dali1Protocol, groups.Standard):
    def exec(self): self._send(gear.Up)


class down(Dali1Protocol, groups.Standard):
    def exec(self): self._send(gear.Down)


class step_up(Dali1Protocol, groups.Standard):
    def exec(self): self._send(gear.StepUp)


class step_down(Dali1Protocol, groups.Standard):
    def exec(self): self._send(gear.StepDown)


class recall_max_level(Dali1Protocol, groups.Standard):
    def exec(self): self._send(gear.RecallMaxLevel)


class recall_min_level(Dali1Protocol, groups.Standard):
    def exec(self): self._send(gear.RecallMinLevel)


class step_down_and_off(Dali1Protocol, groups.Standard):
    def exec(self): self._send(gear.StepDownAndOff)


class on_and_step_up(Dali1Protocol, groups.Standard):
    def exec(self): self._send(gear.OnAndStepUp)


class go_to_last_active_level(Dali1Protocol, groups.Standard):
    def exec(self): self._send(gear.GoToLastActiveLevel)


class go_to_scene(Dali1Protocol, groups.Standard):
    opt_params = {SCENE_NUMBER: 15}
    def exec(self): self._send(gear.GoToScene)


class reset(Dali1Protocol, groups.Standard):
    def exec(self): self._send(gear.Reset)


class set_operating_mode(Dali1Protocol, groups.Standard):
    opt_params = {OPERATING_MODE: 255}
    def exec(self): self._send(gear.SetOperatingMode)


class set_max_level(Dali1Protocol, groups.Standard):
    opt_params = {LEVEL: 254}
    def exec(self): self._send(gear.SetMaxLevel)


class set_min_level(Dali1Protocol, groups.Standard):
    opt_params = {LEVEL: 80}
    def exec(self): self._send(gear.SetMinLevel)


class set_system_failure_level(Dali1Protocol, groups.Standard):
    opt_params = {LEVEL: 80}
    def exec(self): self._send(gear.SetSystemFailureLevel)


class set_power_on_level(Dali1Protocol, groups.Standard):
    opt_params = {LEVEL: 80}
    def exec(self): self._send(gear.SetPowerOnLevel)


class set_fade_time(Dali1Protocol, groups.Standard):
    opt_params = {FADE_TIME: 2}
    def exec(self): self._send(gear.SetFadeTime)


class set_fade_rate(Dali1Protocol, groups.Standard):
    opt_params = {FADE_RATE: 3}
    def exec(self): self._send(gear.SetFadeRate)


class set_scene(Dali1Protocol, groups.Standard):
    opt_params = {SCENE_NUMBER: 15, LEVEL: 254}
    def exec(self): [self._send(gear.SetScene) for i in range(0, int(len(self.params) / 2))]


class remove_from_scene(Dali1Protocol, groups.Standard):
    def exec(self): [self._send(gear.RemoveFromScene) for i in range(0, len(self.params))]


class add_to_group(Dali1Protocol, groups.Standard):
    opt_params = {GROUP_IDs: '1 2 15'}
    def exec(self): [self._send(gear.AddToGroup) for i in range(0, len(self.params))]


class remove_from_group(Dali1Protocol, groups.Standard):
    opt_params = {GROUP_IDs: '1 2 15'}
    def exec(self): [self._send(gear.RemoveFromGroup) for i in range(0, len(self.params))]


class set_short_address(Dali1Protocol, groups.Standard):
    opt_params = {SHORT_ADDRESS: 'D63'}

    def exec(self):
        try:
            self.set_short_address(self.next_param, not settings.conf.quiescent, not settings.conf.quiescent)
        except errors.AddressAlreadyInUse as ex:
            logger.log.exception(ex)
            swap_short_addresses.help(desc=False, example=False)


class query_status(Dali1Protocol, groups.Standard):
    def exec(self): return self._send(gear.QueryStatus)


class query_version_number(Dali1Protocol, groups.Standard):
    def exec(self):
        res = self._send(gear.QueryVersionNumber).value.as_integer
        return f'{res >> 2}.{res & 0x03}'


class query_physical_minimum(Dali1Protocol, groups.Standard):
    def exec(self): return self._send(gear.QueryPhysicalMinimum).value.as_integer


class query_operating_mode(Dali1Protocol, groups.Standard):
    def exec(self): return self._send(gear.QueryOperatingMode).value.as_integer


class query_actual_level(Dali1Protocol, groups.Standard):
    def exec(self): return self._send(gear.QueryActualLevel).value.as_integer


class query_max_level(Dali1Protocol, groups.Standard):
    def exec(self): return self._send(gear.QueryMaxLevel).value.as_integer


class query_min_level(Dali1Protocol, groups.Standard):
    def exec(self): return self._send(gear.QueryMinLevel).value.as_integer


class query_power_on_level(Dali1Protocol, groups.Standard):
    def exec(self): return self._send(gear.QueryPowerOnLevel).value.as_integer


class query_system_failure_level(Dali1Protocol, groups.Standard):
    def exec(self): return self._send(gear.QuerySystemFailureLevel).value.as_integer


class query_fade_time_and_rate(Dali1Protocol, groups.Standard):
    def exec(self):
        res = self._send(gear.QueryFadeTimeFadeRate)
        return {V.FADE_TIME: res.fade_time, V.FADE_RATE: res.fade_rate}


class query_fade_time(Dali1Protocol, groups.Standard):
    VARIANTS = (0.0, 0.7, 1.0, 1.4, 2.0, 2.8, 4.0, 5.7, 8.0, 11.3, 16.0, 22.6, 32.0, 45.3, 64.0, 90.5)
    def exec(self): return self._send(gear.QueryFadeTimeFadeRate).fade_time


class query_fade_rate(Dali1Protocol, groups.Standard):
    VARIANTS = (None, 358, 253, 179, 127, 89.4, 63.3, 44.7, 31.6, 22.4, 15.8, 11.2, 7.9, 5.6, 4.0, 2.8)
    def exec(self): return self._send(gear.QueryFadeTimeFadeRate).fade_rate


class query_scene_level(Dali1Protocol, groups.Standard):
    opt_params = {SCENE_NUMBER: 15}

    def exec(self):
        scenes = {}

        if not self.params:
            self._params = list(self._args[len(self.def_params):]) + list(range(0, 16))

        for scene in self.params.copy():
            try:
                scenes[scene] = self._send(gear.QuerySceneLevel).value.as_integer
            except BaseException as ex:
                scenes[scene] = str(ex)

        return scenes


class store_actual_level(Dali1Protocol, groups.Standard):
    def exec(self):
        self._send(ext_dali.StoreActualLevelInDTR0)
        return self._read_dtr(0).value.as_integer


class query_long_address(Dali1Protocol, groups.Standard):
    def exec(self):
        return {
            'high': self._send(gear.QueryRandomAddressH).value.as_integer,
            'middle': self._send(gear.QueryRandomAddressM).value.as_integer,
            'low': self._send(gear.QueryRandomAddressL).value.as_integer,
        }


""" https://kiwi.throne.systems/index.php/Tron/Strata/Hardware/Rapida/Resh/Command/Dali#Extended """


class query_dimming_curve(Dali1Protocol, groups.Extended):
    def exec(self):
        res = self._send(led.QueryDimmingCurve)
        variants = {0: 'logarithmic', 1: 'linear'}
        return {V.TYPE: res.value.as_integer, V.DESC: variants[res.value.as_integer]}


class select_dimming_curve(Dali1Protocol, groups.Extended):
    opt_params = {DIMMING_CURVE: 0}

    def exec(self): self._send(led.SelectDimmingCurve)


class set_color_temperature(Dali1Protocol, groups.ExtendedColour):
    opt_params = {TEMPERATURE: 1000}

    def exec(self):
        mirek = int(1000000 / max(1000, min(20000, self.next_param)))
        self._params = [int(b) for b in mirek.to_bytes(2, 'little')]
        self._send(cc.SetTemperature)
        self._send(cc.Activate)


class set_color_rgb(Dali1Protocol, groups.ExtendedColour):
    opt_params = {RED: 254, GREEN: 254, BLUE: 254}

    def exec(self):
        self._send(cc.SetRgb)
        self._send(cc.Activate)


class set_color_waf(Dali1Protocol, groups.ExtendedColour):
    opt_params = {WHITE: 254, AMBER: 254, FREE_COLOR: 254}

    def exec(self):
        self._send(cc.SetWaf)
        self._send(cc.Activate)


class _set_color_limit(Dali1Protocol, groups.ExtendedColour):
    _type = None
    opt_params = {TEMPERATURE: 4567}

    def exec(self):
        mirek = int(1000000 / max(1000, min(20000, self.next_param)))
        self._params = [int(b) for b in mirek.to_bytes(2, 'little')] + [self._type]
        self._send(cc.StoreColourTemperatureLimit)


class set_color_coolest(_set_color_limit):
    _type = cc.StoreColourTemperatureLimit.COOLEST


class set_color_warmest(_set_color_limit):
    _type = cc.StoreColourTemperatureLimit.WARMEST


class set_color_physical_coolest(_set_color_limit):
    _type = cc.StoreColourTemperatureLimit.PHYSICAL_COOLEST


class set_color_physical_warmest(_set_color_limit):
    _type = cc.StoreColourTemperatureLimit.PHYSICAL_WARMEST


class _query_color_value(Dali1Protocol, groups.ExtendedColour):
    value_type = None

    def exec(self):
        self._params = [self.value_type]
        res = self._send(cc.QueryColourValue).value.as_integer
        if res != 0xff:
            b0 = self._read_dtr(0).value.as_integer
            b1 = self._read_dtr(1).value.as_integer
            return int(1000000 / int.from_bytes([b0, b1], 'little'))
        return res


class query_color_temperature(_query_color_value):
    value_type = cc.QueryColourValue.TEMPERATURE


class query_color_coolest(_query_color_value):
    value_type = cc.QueryColourValue.COOLEST


class query_color_warmest(_query_color_value):
    value_type = cc.QueryColourValue.WARMEST


class query_color_physical_coolest(_query_color_value):
    value_type = cc.QueryColourValue.PHYSICAL_COOLEST


class query_color_physical_warmest(_query_color_value):
    value_type = cc.QueryColourValue.PHYSICAL_WARMEST


class query_color_rgb(_query_color_value):
    def exec(self):
        color = {}
        for value_type, value_name in (
                (cc.QueryColourValue.RED, 'red'),
                (cc.QueryColourValue.GREEN, 'green'),
                (cc.QueryColourValue.BLUE, 'blue'),
                (cc.QueryColourValue.WHITE, 'white'),
                (cc.QueryColourValue.AMBER, 'amber'),
                (cc.QueryColourValue.FREE_COLOUR, 'freeColor'),
        ):
            self._params = [value_type]
            value = self._send(cc.QueryColourValue).value.as_integer
            if value != 255:
                color[value_name] = value

        return color


class query_color_status(Dali1Protocol, groups.ExtendedColour):
    def exec(self): return self._send(cc.QueryColourStatus)


class query_color_type_features(Dali1Protocol, groups.ExtendedColour):
    def exec(self): return self._send(cc.QueryColourTypeFeatures).status


class query_color_rgbwaf_control(Dali1Protocol, groups.ExtendedColour):
    def exec(self): return self._send(cc.QueryRgbwafControl).status


class query_color_assigned(Dali1Protocol, groups.ExtendedColour):
    def exec(self): return self._send(cc.QueryAssignedColour).status


class _set_switch_value(Dali1Protocol, groups.ExtendedSwitch):
    opt_params = {'value': 127}


class _query_switch_value(Dali1Protocol, groups.ExtendedSwitch):
    pass


class set_up_switch_on(_set_switch_value):
    def exec(self): self._send(sf.SetUpSwitchOnThreshold)


class set_up_switch_off(_set_switch_value):
    def exec(self): self._send(sf.SetUpSwitchOffThreshold)


class set_down_switch_on(_set_switch_value):
    def exec(self): self._send(sf.SetDownSwitchOnThreshold)


class set_down_switch_off(_set_switch_value):
    def exec(self): self._send(sf.SetDownSwitchOffThreshold)


class set_error_hold_off_time(_set_switch_value):
    def exec(self): self._send(sf.SetErrorHoldOffTime)


class query_up_switch_on(_query_switch_value):
    def exec(self): return self._send(sf.QueryUpSwitchOnThreshold).value.as_integer


class query_up_switch_off(_query_switch_value):
    def exec(self): return self._send(sf.QueryUpSwitchOffThreshold).value.as_integer


class query_down_switch_on(_query_switch_value):
    def exec(self): return self._send(sf.QueryDownSwitchOnThreshold).value.as_integer


class query_down_switch_off(_query_switch_value):
    def exec(self): return self._send(sf.QueryDownSwitchOffThreshold).value.as_integer


class query_error_hold_off_time(_query_switch_value):
    def exec(self): return self._send(sf.QueryErrorHoldOffTime).value.as_integer


class query_virtual_level(_query_switch_value):
    def exec(self): return store_actual_level(self.module, self.channel, tools.pack_address(self.address)).exec()


class query_features_status(Dali1Protocol, groups.Extended):
    def exec(self): return self._send(cc.QueryGearFeaturesStatus)


"""https://kiwi.throne.systems/index.php/Tron/Strata/Hardware/Rapida/Resh/Command/Dali#Wrappers"""


class remove_short_address(Dali1Protocol, groups.Wrappers):
    def exec(self):
        self.send(gear.DTR0(0xff))
        self.send(gear.SetShortAddress(self.address))


class query_groups(Dali1Protocol, groups.Wrappers):
    def exec(self):
        result = tools.bits_indexes(
            self._send(gear.QueryGroupsZeroToSeven)
                .value.as_integer, start_from=0
        ) + tools.bits_indexes(
            self._send(gear.QueryGroupsEightToFifteen)
                .value.as_integer, start_from=8
        )
        return result


class query_device_types(Dali1Protocol, groups.Wrappers):
    def exec(self): return self.query_device_types()


class read_firmware_version(Dali1Protocol, groups.Wrappers):
    def exec(self):
        result = b''
        for addr in range(0x9, 0xb):
            result += self.read_memory_bank(0, addr)
        return '.'.join(str(b) for b in result)


class read_hardware_version(Dali1Protocol, groups.Wrappers):
    def exec(self):
        result = b''
        for addr in range(0x13, 0x15):
            result += self.read_memory_bank(0, addr)
        return '.'.join(str(b) for b in result)


class read_gtin(Dali1Protocol, groups.Wrappers):
    def exec(self):
        result = b''
        for addr in range(0x03, 0x09):
            result += self.read_memory_bank(0, addr)
        return ''.join(hex(b)[2:].rjust(2, '0') for b in result).upper()


class read_gtin_oem(Dali1Protocol, groups.Wrappers):
    def exec(self):
        result = b''
        for addr in range(0x03, 0x09):
            result += self.read_memory_bank(1, addr)
        return ''.join(hex(b)[2:].rjust(2, '0') for b in result).upper()


class read_serial(Dali1Protocol, groups.Wrappers):
    def exec(self):
        result = b''
        for addr in range(0x0b, 0x13):
            result += self.read_memory_bank(0, addr)
        return ''.join(hex(b)[2:].rjust(2, '0') for b in result).upper()


class read_serial_oem(Dali1Protocol, groups.Wrappers):
    def exec(self):
        result = b''
        for addr in range(0x9, 0x11):
            result += self.read_memory_bank(1, addr)
        return ''.join(hex(b)[2:].rjust(2, '0') for b in result).upper()


"""https://kiwi.throne.systems/index.php/Tron/Strata/Hardware/Rapida/Resh/Command/Dali#Wizards"""


class identify(Dali1Protocol, groups.Wizards):
    var_params = {TIME: 9, LEVEL: 127}
    __stopped = False
    __repeater = None
    __delay = 10
    _need_level = None

    """Не доступно для групп"""
    def exec(self):
        if isinstance(self.address, dali.address.Group):
            raise errors.DeviceAddressError()

        _broadcast = isinstance(self.address, dali.address.Broadcast)
        _address = None

        if not _broadcast:
            _address = self.address.address

        if len(self.params) > 0:
            timeout = self.next_param
            if not isinstance(timeout, int):
                raise errors.BadParameter()
        else:
            timeout = 3600

        if self.has_next:
            self._need_level = self.next_param
            if self._need_level not in range(0, 255):
                raise errors.BadParameter()
        else:
            try:
                self._need_level = self.send(gear.QueryActualLevel(self.address)).value.as_integer
            except:
                pass

        self.send(gear.Initialise(_broadcast, _address))
        self.send(gear.IdentifyDevice(self.address))

        if timeout > 0:
            self.__start()

            if threading.current_thread() is threading.main_thread():
                logger.log.question(loc.IDENTIFY_QUESTION, timeout)
            else:
                sleep(timeout)

        self.stop()

    def stop(self):
        self.__stopped = True
        if self.__repeater:
            self.__repeater.cancel()
            self.__repeater.join()
            self.__repeater = None
            self.send(gear.Terminate())
            if isinstance(self._need_level, int):
                self.send(gear.DAPC(self.address, self._need_level))

    def __start(self):
        if self.__stopped is False:
            self.send(gear.RecallMaxLevel(self.address))
            self.__repeater = threading.Timer(self.__delay, function=self.__start)
            self.__repeater.start()


class discovery(Dali1Protocol, groups.Wizards):
    var_params = {TIME: 9, LEVEL: 127}
    _need_level = None
    __stopped = False
    __repeater = None
    __delay = 0.5

    def exec(self):
        if self.has_next:
            timeout = self.next_param
            if not isinstance(timeout, int):
                raise errors.BadParameter()
        else:
            timeout = 3600

        if self.has_next:
            self._need_level = self.next_param
            if self._need_level not in range(0, 255):
                raise errors.BadParameter()
        else:
            try:
                self._need_level = self.send(gear.QueryActualLevel(self.address)).value.as_integer
            except:
                pass

        if timeout > 0:
            self.__max()
            if threading.current_thread() is threading.main_thread():
                logger.log.question(loc.IDENTIFY_QUESTION, timeout)
            else:
                sleep(timeout)

        self.stop()

    def stop(self):
        self.__stopped = True
        if self.__repeater:
            self.__repeater.cancel()
            self.__repeater.join()
            self.__repeater = None
            if isinstance(self._need_level, int):
                self.send(gear.DAPC(self.address, self._need_level))

    def __next(self, command, next_step_func):
        if self.__stopped is False:
            self.send(command)
            self.__repeater = threading.Timer(self.__delay, function=next_step_func)
            self.__repeater.start()

    def __max(self):
        self.__next(gear.RecallMaxLevel(self.address), self.__min)

    def __min(self):
        self.__next(gear.Off(self.address), self.__max)


class swap_short_addresses(Dali1Protocol, groups.Wizards):
    def_params = {MODULE: 1, CHANNEL: 2}
    opt_params = {ADDRESS1: 'D0', ADDRESS2: 'D63'}

    def _write_short_address(self, long_address, short_address: dali.address.Short):
        self.send(gear.SetSearchAddrH(long_address[0]))
        self.send(gear.SetSearchAddrM(long_address[1]))
        self.send(gear.SetSearchAddrL(long_address[2]))

        self.send(gear.ProgramShortAddress(short_address.address))

        try:
            r = self.send(gear.QueryShortAddress()).value.as_integer >> 1
            if r != short_address.address:
                raise errors.CustomError(
                    'Response of address(%s) is not equal to value set(%s)' % (r, short_address.address)
                )
        finally:
            self.send(gear.Withdraw())

    def exec(self):
        self._address = tools.parse_address(self.next_param)
        raw_new_address = self.next_param
        new_short = tools.parse_address(raw_new_address)
        if not isinstance(new_short, dali.address.Short):
            raise errors.BadParameter()

        try:
            self.set_short_address(raw_new_address, need_warning=False)
        except errors.AddressAlreadyInUse:
            old_short = self.address

            self.send(gear.Terminate())
            self.send(gear.Initialise(broadcast=True))
            self.send(gear.Randomise())
            sleep(0.3)

            try:
                old_long = (
                    self.send(gear.QueryRandomAddressH(old_short)).value.as_integer,
                    self.send(gear.QueryRandomAddressM(old_short)).value.as_integer,
                    self.send(gear.QueryRandomAddressL(old_short)).value.as_integer
                )

                new_long = (
                    self.send(gear.QueryRandomAddressH(new_short)).value.as_integer,
                    self.send(gear.QueryRandomAddressM(new_short)).value.as_integer,
                    self.send(gear.QueryRandomAddressL(new_short)).value.as_integer
                )

                self._write_short_address(old_long, new_short)
                self._write_short_address(new_long, old_short)
            finally:
                self.send(gear.Terminate())


class _line_commands(Dali1Protocol, groups.Wizards):
    def exec(self): raise NotImplementedError()

    def_params = {MODULE: 1, CHANNEL: 2}
    opt_params = {MODE: MODE_CROSS}
    var_params = {START_ADDRESS: 0}
    __bus = None
    __is_cross_addressing = None
    __start_index = None

    @property
    def _bus(self):
        if self.__bus is None:
            self.__bus = Bus1(self, self._is_cross_addressing, self._start_index, step_handler=self.on_step)
        return self.__bus

    @property
    def _is_cross_addressing(self):
        if self.__is_cross_addressing is None:
            cross_addr = str(self.next_param).upper()
            if cross_addr == MODE_NOCROSS:
                self.__is_cross_addressing = False
            elif cross_addr == MODE_CROSS:
                self.__is_cross_addressing = True
            else:
                raise errors.BadParameter()

        return self.__is_cross_addressing

    @property
    def _start_index(self):
        if self.__start_index is None:
            if len(self.params) > 0:
                index = self.next_param
                if isinstance(index, int):
                    self.__start_index = index
                else:
                    raise errors.BadParameter()
            else:
                self.__start_index = 0

        return self.__start_index


class scan_line(_line_commands):
    def exec(self):
        scan = {}
        for addr in range(0, 64):
            self._address = dali.address.Short(addr)
            self._device_types = None
            try:
                scan[addr] = self.device_types
            except errors.DeviceNotAnswered:
                continue
        return scan

    @property
    def _is_cross_addressing(self): return False

    @property
    def _start_index(self): return 0


class test_line(Dali1Protocol, groups.Wizards):
    DEF_COUNT = 10
    DEF_COMMAND = gear.QueryActualLevel

    def_params = {MODULE: 1, CHANNEL: 2}
    var_params = {COUNT: DEF_COUNT, COMMAND: DEF_COMMAND.__name__, ADDRESSES: 'D0,D64', SORT_BY_ADDR: ''}

    def _step_handler(self, progress, *args, **kwargs):
        if progress < 100:
            sys.stdout.write('\rProgress: {}%'.format(progress))
        else:
            sys.stdout.write('\r' + ' ' * 14 + '\r')

        sys.stdout.flush()

    def exec(self):
        sort_by_addr = False
        if SORT_BY_ADDR in self._args:
            sort_by_addr = True
            self._args.remove(SORT_BY_ADDR)

        count = self.DEF_COUNT
        command = self.DEF_COMMAND

        if self.has_next:
            if isinstance(self.next_param_recursive, int):
                count = self.next_param
                if count < 1:
                    raise errors.BadParameter('Количество запросов должно быть больше 0')

        if self.has_next:
            commands = {
                cmd.__name__: cmd
                for cmd in gear._StandardCommand.__subclasses__() + led._LEDCommand.__subclasses__()
                if issubclass(cmd, gear._StandardCommand) and cmd._response is not None
            }

            if self.next_param_recursive in commands:
                command = commands[self.next_param]

        if self.has_next:
            try:
                addresses = sorted(
                    [
                        tools.parse_address(a.strip())
                        for a in self.next_param.split(',')
                    ],
                    key=lambda a: a.address
                )

            except:
                raise errors.BadParameter('Ошибка парсинга адресов')
        else:
            addresses = [dali.address.Short(i) for i in range(0, 64)]

        result = {}
        step = 0
        self.start_quiescent_mode()
        for i in range(0, count):
            if self.has_break:
                break
            for addr in addresses:
                if self.has_break:
                    break

                step += 1

                result.setdefault(addr.address, 0)
                try:
                    self._address = addr
                    self._send(command)
                    result[addr.address] += 1
                except BaseException:
                    continue
                finally:
                    if not logger.log.is_json:
                        self.on_step(min(100, max(1, int(step / (count * len(addresses)) * 100))))

        self.stop_quiescent_mode()

        result = {a: '{}%'.format(int(v * 100 / count)) for a, v in result.items()}

        if not sort_by_addr:
            result = dict(sorted(result.items(), key=lambda item: item[1], reverse=True))

        return result


class test_device(Dali1Protocol, groups.Wizards):
    class Scenes:
        def __init__(self, m4a):
            self.__owner_address = m4a
            self._levels = [None for i in range(0, 16)]

        def __setitem__(self, index, level):
            self._levels[index] = None
            set_scene(*self.__owner_address, index, level).exec()

        def __getitem__(self, index):
            if 0 <= index < 16:
                if self._levels[index] is None:
                    self._levels[index] = query_scene_level(*self.__owner_address, index).exec()[index]

                return self._levels[index]

            raise KeyError(index)

        def __iter__(self):
            for i in range(0, 16):
                yield self[i]

        @property
        def as_dict(self):
            return {i: self[i] for i in range(0, 16)}

        @staticmethod
        def as_params(values: dict) -> list:
            result = []
            for i, l in values.items():
                result.append(i)
                result.append(l)

            return result

    def __init__(self, *args, **kwargs):
        super(test_device, self).__init__(*args, **kwargs)

        self.__m4a = self.module, self.channel, 'D%s' % self.address.address

        self._level = None
        self._min_level = None
        self._max_level = None
        self._ph_min_level = None
        self._fade_time = None
        self._fade_rate = None
        self._groups = None
        self._scenes = None

    @property
    def ph_min_level(self):
        if self._ph_min_level is None:
            self._ph_min_level = query_physical_minimum(*self.__m4a).exec()
        return self._ph_min_level

    @property
    def min_level(self):
        if self._min_level is None:
            self._min_level = max(self.ph_min_level, query_min_level(*self.__m4a).exec())
        return self._min_level

    @min_level.setter
    def min_level(self, value):
        self._min_level = None
        if self._level is not None and self._level <= value:
            self._level = None
        set_min_level(*self.__m4a, value).exec()

    @property
    def max_level(self):
        if self._max_level is None:
            self._max_level = max(self.min_level, query_max_level(*self.__m4a).exec())

        return self._max_level

    @max_level.setter
    def max_level(self, value):
        self._max_level = None
        if self._level is not None and self._level >= value:
            self._level = None
        set_max_level(*self.__m4a, value).exec()

    @property
    def level(self):
        if self._level is None:
            for i in range(0, 2):
                self._level = query_actual_level(*self.__m4a).exec()
                if self._level != 255:
                    break

        return self._level

    @level.setter
    def level(self, value):
        self._level = None
        set_actual_level(*self.__m4a, value).exec()

    @property
    def fade_time(self):
        if self._fade_time is None:
            self._fade_time = query_fade_time(*self.__m4a).exec()
        return self._fade_time

    @fade_time.setter
    def fade_time(self, value):
        set_fade_time(*self.__m4a, value).exec()
        self._fade_time = value

    @property
    def fade_rate(self):
        if self._fade_rate is None:
            self._fade_rate = query_fade_time(*self.__m4a).exec()
        return self._fade_rate

    @fade_rate.setter
    def fade_rate(self, value):
        set_fade_rate(*self.__m4a, value).exec()
        self._fade_rate = value

    @property
    def status(self):
        return query_status(*self.__m4a).exec().status

    @property
    def groups(self):
        if self._groups is None:
            self._groups = query_groups(*self.__m4a).exec()

        return self._groups

    @property
    def scenes(self) -> Scenes:
        if self._scenes is None:
            self._scenes = test_device.Scenes(self.__m4a)

        return self._scenes

    @scenes.setter
    def scenes(self, value):
        for index, level in value.items():
            self.scenes[index] = level

    def off(self):
        self._level = None
        off(*self.__m4a).exec()

    def recall_min_level(self):
        self._level = None
        recall_min_level(*self.__m4a).exec()
        sleep(0.2)

    def recall_max_level(self):
        self._level = None
        recall_max_level(*self.__m4a).exec()
        sleep(0.3)

    def up(self):
        self._level = None
        up(*self.__m4a).exec()
        sleep(0.3)

    def down(self):
        self._level = None
        down(*self.__m4a).exec()
        sleep(0.3)

    def add_to_groups(self, *gr):
        self._groups = None
        add_to_group(*self.__m4a, *gr)

    def remove_from_groups(self, *gr):
        self._groups = None
        remove_from_group(*self.__m4a, *gr)

    def _check_simple(self, variable, setter, getter, value, expected=None):
        if variable is None:
            if setter:
                variable = setter.__name__
            elif getter:
                variable = getter.__name__
        elif not isinstance(variable, str):
            variable = variable.__name__

        if expected is None:
            has_exp = False
            expected = value
        else:
            has_exp = True

        if setter:
            if value is None:
                setter(*self.__m4a).exec()
            elif isinstance(value, list):
                setter(*self.__m4a, *value).exec()
            else:
                setter(*self.__m4a, value).exec()

            if setter in (recall_min_level, recall_max_level, off, set_max_level):
                sleep(0.2)
        if getter:
            response = getter(*self.__m4a).exec()

            if getter == query_actual_level and response == 255:
                sleep(0.5)
                response = getter(*self.__m4a).exec()

            if expected != response:
                self.on_step(variable, 'set: "%s"; get "%s"' % (expected, response) + ', expected "%s"' % expected if has_exp else '')

            return response

    def _check_fade_time(self):
        has_level = self.level
        has_value = self.fade_time

        for i in range(0, 16):
            self._check_simple(V.FADE_TIME, set_fade_time, query_fade_time, i)

            if i > 0:
                if i <= 6:
                    self.off()

                self.level = 254

                sleep(0.2)
                if 'fade ready' not in self.status:
                    self.on_step(V.FADE_TIME, 'Fade flag is not set when "DAPC" with %s' % i)

                start = time()
                levels = []
                while time() < start + min(query_fade_time.VARIANTS[6], query_fade_time.VARIANTS[i]):
                    levels.append(query_actual_level(*self.__m4a).exec())

                levels = set(levels)
                if len(levels) < 3:
                    self.on_step(V.FADE_TIME, 'Fading has only %s step with ft=%s' % (len(levels), i))

                if i >= 6:
                    self.off()

                sleep(0.2)
                if 'fade ready' in self.status:
                    self.on_step(V.FADE_TIME, 'Fade flag is not unset after "DAPC" with %s' % i)

        self.level = has_level
        self.fade_time = has_value

    def _check_fade_rate(self):
        has_level = self.level
        has_value = self.fade_rate
        self.recall_min_level()
        level = self.level

        self._check_simple(V.FADE_RATE, set_fade_rate, query_fade_rate, 0, expected=1)

        for i in range(1, 16):
            self._check_simple(V.FADE_RATE, set_fade_rate, query_fade_rate, i)

            fr = int(query_fade_rate.VARIANTS[i] * 0.2)

            self.up()
            new_level = self.level
            if new_level != level + fr:
                self.on_step(V.FADE_RATE, '"Up" rated level for %s but expected %s' % (new_level-level, fr))
            level = new_level

            self.down()
            new_level = self.level
            if new_level != level - fr:
                self.on_step(V.FADE_RATE, '"Down" rated level for %s but expected %s' % (level - new_level, fr))
            level = new_level

        self.level = has_level
        self.fade_rate = has_value

    def _check_levels(self):
        has_level = self.level
        try:
            min_level = self.min_level
        except BaseException as ex:
            self.on_step(V.MIN_LEVEL, ex)
            min_level = 1

        try:
            max_level = self.max_level
        except BaseException as ex:
            self.on_step(V.MAX_LEVEL, ex)
            max_level = 254

        has_fade_time = self.fade_time
        self.fade_time = 0

        self._check_simple(None, off, query_actual_level, None, expected=0)

        # MinLevel
        self._check_simple(None, recall_min_level, query_actual_level, None, expected=min_level)
        self._check_simple(V.MIN_LEVEL, set_min_level, query_min_level, max_level)
        self._check_simple(V.ACTUAL_LEVEL, None, query_actual_level, None, expected=max_level)
        self._check_simple(V.MIN_LEVEL, set_min_level, query_min_level, min_level+1)
        self._check_simple(V.ACTUAL_LEVEL, set_actual_level, query_actual_level, min_level, expected=min_level+1)
        self._check_simple(V.MIN_LEVEL, set_min_level, query_min_level, min_level)
        self._check_simple(V.ACTUAL_LEVEL, set_actual_level, query_actual_level, min_level)

        # MaxLevel
        self._check_simple(None, recall_max_level, query_actual_level, None, expected=max_level)
        self._check_simple(V.MAX_LEVEL, set_max_level, query_max_level, min_level)
        self._check_simple(V.ACTUAL_LEVEL, None, query_actual_level, None, expected=min_level)
        self._check_simple(V.MAX_LEVEL, set_max_level, query_max_level, max_level-1)
        self._check_simple(V.ACTUAL_LEVEL, set_actual_level, query_actual_level, max_level, expected=max_level-1)
        self._check_simple(V.MAX_LEVEL, set_max_level, query_max_level, max_level)
        self._check_simple(V.ACTUAL_LEVEL, set_actual_level, query_actual_level, max_level)

        # DAPC
        for level in range(1, 255):
            self._check_simple(V.ACTUAL_LEVEL, set_actual_level, query_actual_level, min(max_level, max(min_level, level)))
        for level in range(253, 0, -1):
            self._check_simple(V.ACTUAL_LEVEL, set_actual_level, query_actual_level, min(max_level, max(min_level, level)))

        self._check_simple(V.ACTUAL_LEVEL, set_actual_level, query_actual_level, 0)

        # STEPS
        self._check_simple(None, on_and_step_up, query_actual_level, None, expected=min_level)
        for level in range(min_level, max_level):
            self._check_simple(None, step_up, query_actual_level, None, expected=min(max_level, max(min_level, level+1)))
        for level in range(max_level, min_level, -1):
            self._check_simple(None, step_down, query_actual_level, None, expected=min(max_level, max(min_level, level-1)))

        self._check_simple(None, step_down_and_off, query_actual_level, None, expected=0)

        self.level = max_level
        self.off()
        self._check_simple(None, go_to_last_active_level, query_actual_level, None, max_level)

        self.level = has_level
        self.fade_time = has_fade_time

    def _check_groups(self):
        has_level = self.level
        has_groups = self.groups.copy()

        self._check_simple(None, remove_from_group, query_groups, [i for i in range(0, 16)], expected=[])
        self._check_simple(None, add_to_group, query_groups, [i for i in range(0, 16)])

        self.recall_max_level()
        level = self.level
        for group in range(0, 16):
            step_down(self.module, self.channel, 'G%s' % group).exec()
            response = query_actual_level(*self.__m4a).exec()

            if level - 1 != response:
                self.on_step(V.GROUPS, 'send %s to G%s but level is %s' % (step_down.__name__, group, response))

            level = response

        self._check_simple(None, remove_from_group, query_groups, [i for i in range(0, 16)], expected=[])
        self._check_simple(None, add_to_group, query_groups, has_groups)

        self.level = has_level

    def _check_scenes(self):
        _EMPTY = {i: 255 for i in range(0, 16)}

        has_scenes = self.scenes.as_dict
        has_level = self.level
        has_fade_time = self.fade_time

        self.fade_time = 0
        self.recall_min_level()

        self._check_simple(None, remove_from_scene, query_scene_level, [i for i in range(0, 16)], expected=_EMPTY)
        scenes = {i: min(self.max_level, self.min_level+i) for i in range(3, 16)}
        scenes[0] = 0
        scenes[1] = 255
        scenes[2] = self.max_level
        scenes = self._check_simple(None, set_scene, query_scene_level, self.Scenes.as_params(scenes), expected=scenes)

        for index, level in scenes.items():
            self._check_simple(None, go_to_scene, query_actual_level, index, expected=level if level < 255 else 0)

        self._check_simple(None, set_scene, query_scene_level, self.Scenes.as_params(has_scenes), expected=has_scenes)

        self.scenes = has_scenes
        self.level = has_level
        self.fade_time = has_fade_time

    def exec(self):
        self.start_quiescent_mode()
        try:
            try:
                types = self.device_types
                self.on_step(DEVICE_TYPE, types)
            except BaseException as ex:
                self.on_step(DEVICE_TYPE, ex)

            self._check_fade_time()
            self._check_fade_rate()
            self._check_levels()
            self._check_groups()
            self._check_scenes()
        finally:
            self.stop_quiescent_mode()

    def on_step(self, variable, message):
        print('%s: %s' % (variable, message))


class construct_line(_line_commands):
    def exec(self): return self._bus.assign_short_addresses(False, broadcast=True)


class extend_line(_line_commands):
    def exec(self): return self._bus.assign_short_addresses(True, broadcast=False)


class fix_double_addresses(_line_commands):
    def exec(self): return self._bus.address_resolution(broadcast=True)


class dump(Dali1Protocol, groups.Wizards):
    queries = {
        V.SERIAL: dict(get=read_serial),
        V.GTIN: dict(get=read_gtin),
        V.HARDWARE: dict(get=read_hardware_version),
        V.FIRMWARE: dict(get=read_firmware_version),
        V.GTIN_OEM: dict(get=read_gtin_oem),
        V.SERIAL_OEM: dict(get=read_serial_oem),
        V.GROUPS: dict(get=query_groups),
        V.ACTUAL_LEVEL: dict(get=query_actual_level, set=set_actual_level),
        V.FADE_TIME_AND_RATE: dict(get=query_fade_time_and_rate),
        V.FADE_TIME: dict(set=set_fade_time),
        V.FADE_RATE: dict(set=set_fade_rate),
        V.MIN_LEVEL: dict(get=query_min_level, set=set_min_level),
        V.MAX_LEVEL: dict(get=query_max_level, set=set_max_level),
        V.SYSTEM_FAILURE: dict(get=query_system_failure_level, set=set_system_failure_level),
        V.POWER_ON_LEVEL: dict(get=query_power_on_level, set=set_power_on_level),
        V.PHYSICAL_MINIMUM: dict(get=query_physical_minimum),
        V.OPERATING_MODE: dict(get=query_operating_mode, set=set_operating_mode),
        V.DISCOVERY: dict(set=discovery),
        V.IDENTIFY: dict(set=identify),
        V.SCENES: dict(get=query_scene_level),
    }

    extended = {
        6: {
            V.DIMMING_CURVE: dict(get=query_dimming_curve, set=select_dimming_curve),
        },
        7: {
            V.SWITCH_UP_ON: dict(get=query_up_switch_on, set=set_up_switch_on),
            V.SWITCH_UP_OFF: dict(get=query_up_switch_off, set=set_up_switch_off),
            V.SWITCH_DOWN_ON: dict(get=query_down_switch_on, set=set_down_switch_on),
            V.SWITCH_DOWN_OFF: dict(get=query_down_switch_off, set=set_down_switch_off),
            V.VIRTUAL_LEVEL: dict(get=query_virtual_level),
        },
        8: {
            V.CC_COOLEST: dict(get=query_color_coolest, set=set_color_coolest),
            V.CC_WARMEST: dict(get=query_color_warmest, set=set_color_warmest),
            V.CC_PH_COOLEST: dict(get=query_color_physical_coolest),
            V.CC_PH_WARMEST: dict(get=query_color_physical_warmest),
            V.CC_COLOR_TEMPERATURE: dict(get=query_color_temperature, set=set_color_temperature),
        }
    }

    def __init__(self, *args, **kwargs):
        super(dump, self).__init__(*args, **kwargs)
        self._custom_stack = {}

    @property
    def custom_stack(self): return self._custom_stack

    @custom_stack.setter
    def custom_stack(self, stack): self._custom_stack = stack

    def need_key(self, key): return not self._custom_stack or key in self._custom_stack

    def exec(self):
        devices = []
        root_dev = self.address

        to_rise = None

        if self.address.__class__ == dali.address.Broadcast:
            devices = [dali.address.Short(i) for i in range(0, 64)]
        elif self.address.__class__ in (dali.address.Group, dali.address.BroadcastUnaddressed):
            raise errors.CustomError('Can not fill dump for {}'.format(self.address.__class__.__name__))
        else:
            devices.append(self.address)

        result = dict()

        self.start_quiescent_mode()

        for addr in devices:
            if self.has_break:
                break

            self._address = addr
            try:
                types = self.device_types
            except errors.DeviceNotAnswered as ex:
                if isinstance(root_dev, dali.address.Short):
                    to_rise = ex
                    break
            else:
                result[addr.address] = {'types': types}

                self.__on_progress(f'{addr.address}:')
                self.__on_progress(f'   types: {types}')
                self.on_step(addr.address, 'types', types)

                queries = {**self.queries}
                for _type in types:
                    queries.update(self.extended.get(_type, {}))

                for key, query_classes in queries.items():
                    if self.has_break:
                        break

                    if self.need_key(key):
                        query = query_classes.get('get')
                        if query:
                            try:
                                query = query()
                                query._module = self.module
                                query._channel = self.channel
                                query._address = addr
                                result[addr.address][key] = query.exec()
                            except BaseException as ex:
                                result[addr.address][key] = str(ex).strip()
                                self.on_step(addr.address, key, ex)
                            else:
                                self.on_step(addr.address, key, result[addr.address].get(key, 'error'))
                            finally:
                                self.__on_progress(f'   {key}: {result[addr.address].get(key, "error")}')

        self.stop_quiescent_mode()

        if isinstance(to_rise, BaseException):
            raise to_rise
        elif logger.log.is_json:
            return result

    def __on_progress(self, string):
        if not logger.log.is_json:
            print(string)
