import sys
import threading
from time import sleep

import dali.address
from dali.device import general as device

from resh.domains._domain import DaliProtocol
from resh.locales.loc_def import loc
from resh.protocols.dali_utils.bus2 import Bus2
from resh.protocols.dali_utils.external_commands.dali2 import presence_sensor_ext, light_sensor_ext, push_button_ext, \
    dali2_responses as device_ext
from resh.utils import groups, tools, errors, logger, compat, const_values as v, settings
from resh.utils.const import *

BUTTON = 1
PRESENCE = 3
LIGHT = 4

DEF_PARAMS_INSTANCE = {MODULE: 1, CHANNEL: 2, DEVICE_ADDR: 'D3', INSTANCE: 'IT3'}


@compat.add_metaclass(compat.DomainTracker)
class Dali2Protocol(DaliProtocol):
    def_params = {MODULE: 1, CHANNEL: 2, DEVICE_ADDR: 'D3'}
    domain_name = DALI2
    _bytes_count = 3

    def __init__(self, *args):
        super(Dali2Protocol, self).__init__(*args)
        self._instance = None
        self._instance_type = None
        self._instance_count = None

    @property
    def instance(self):
        if not self._instance:
            if len(self._args) < 4:
                raise errors.InstanceAddressError()
            else:
                self._instance = tools.parse_instance(self._args[3])
                if isinstance(self._instance, dali.address.InstanceType):
                    self._instance_type = self._instance._value
        return self._instance

    @property
    def instance_type(self):
        if self._instance_type is None:
            self._instance_type = self._send(device.QueryInstanceType).value.as_integer

        return self._instance_type

    @property
    def instance_count(self):
        if self._instance_count is None:
            self._instance_count = self._send(device.QueryNumberOfInstances).value.as_integer

        return self._instance_count

    def _check_instance(self, *need_types):
        if not isinstance(self.address, dali.address.Short):
            if not isinstance(self.instance, (dali.address.InstanceBroadcast, dali.address.InstanceGroup)):
                if self.instance_type not in need_types:
                    raise errors.UnavailableInstance()

    def _send(self, cmd_class):
        if issubclass(cmd_class, device._StandardDeviceCommand):
            _cmd = cmd_class(self.address)
        elif issubclass(cmd_class, device._StandardInstanceCommand):
            _cmd = cmd_class(self.address, self.instance)
        else:
            _cmd = cmd_class()

        try:
            dtr_cmd = None
            if cmd_class._uses_dtr0 and cmd_class._uses_dtr1:
                _dtr0 = self.next_param
                _dtr1 = self.next_param
                dtr_cmd = device.DTR1DTR0(_dtr0, _dtr1)
            elif cmd_class._uses_dtr1 and cmd_class._uses_dtr2:
                _dtr1 = self.next_param
                _dtr2 = self.next_param
                dtr_cmd = device.DTR2DTR1(_dtr1, _dtr2)
            elif cmd_class._uses_dtr0:
                dtr_cmd = device.DTR0(self.next_param)
            elif cmd_class._uses_dtr1:
                dtr_cmd = device.DTR1(self.next_param)
            elif cmd_class._uses_dtr2:
                dtr_cmd = device.DTR2(self.next_param)
        except:
            raise errors.BadParameter()

        if dtr_cmd:
            self.send(dtr_cmd)

        return self.send(_cmd)

    def write_memory_bank(self, bank_id=None, bank_addr=None, value=None):
        try:
            if bank_id is None:
                bank_id = self.next_param
            if bank_addr is None:
                bank_addr = self.next_param
            if value is None:
                value = self.next_param
        except:
            raise errors.BadParameter()

        self.send(device.EnableWriteMemory(self.address))
        self.send(device.DTR1(bank_id))

        self.send(device.DTR0(0x02))
        self.send(device.WriteMemoryLocation(0x55))

        self.send(device.DTR0(bank_addr))
        resp = self.send(device.WriteMemoryLocation(value))

        self.send(device.DTR0(0x02))
        self.send(device.WriteMemoryLocation(0xFF))

        return resp.value.pack

    def read_memory_bank(self, bank_id=None, bank_addr=None):
        try:
            if bank_id is None:
                bank_id = self.next_param
            if bank_addr is None:
                bank_addr = self.next_param
            dtr = device.DTR1DTR0(bank_id, bank_addr)
        except:
            raise errors.BadParameter()

        self.send(dtr)

        resp = self.send(device.ReadMemoryLocation(self.address))
        return resp.value.pack

    def set_short_address(self, new_address, need_warning=True, raise_if_used=True):
        try:
            new_address = tools.parse_address(new_address).address
        except:
            raise errors.BadParameter()

        try:
            self.send(device.QueryNumberOfInstances(new_address))
        except errors.DeviceNotAnswered:
            pass
        else:
            answer = None
            if need_warning:
                answer = logger.log.question(f'Address {new_address} will be duplicated. Continue? (y/N)', timeout=5)

            if raise_if_used and answer not in ['yes', 'y', '1', 'da']:
                raise errors.AddressAlreadyInUse(new_address)

        self.send(device.DTR0(new_address))
        self.send(device.SetShortAddress(self.address))


"""https://kiwi.throne.systems/index.php/Tron/Strata/Hardware/Rapida/Resh/Command/Dali2#Base"""


class raw_command(Dali2Protocol, groups.Base):
    def_params = {MODULE: 1, CHANNEL: 2}
    opt_params = {DEVICE_ADDR: '0x00', BYTE_INSTANCE: '0x00', BYTE_OPCODE: '0xfe'}
    var_params = {SEND_TWICE: SEND_TWICE}

    def exec(self): self.raw_command(False)


class raw_command_dtr0(Dali2Protocol, groups.Base):
    def_params = {MODULE: 1, CHANNEL: 2}
    opt_params = {BYTE_ADDR: '0x00', BYTE_INSTANCE: '0x00', BYTE_OPCODE: '0xfe', DTR0_VALUE: '255'}
    var_params = {SEND_TWICE: SEND_TWICE}

    def exec(self):
        self.send(device.DTR0(self.indexed_param(3)))
        self.raw_command(False)


class raw_query(Dali2Protocol, groups.Base):
    def_params = {MODULE: 1, CHANNEL: 2}
    opt_params = {DEVICE_ADDR: '0x01', BYTE_INSTANCE: '0x00', BYTE_OPCODE: '0xA0'}

    def exec(self): return self.raw_command(True)


class raw_read_memory(Dali2Protocol, groups.Base):
    opt_params = {MEMORY_BANK: 0, MEMORY_LOCATION: 8}
    def exec(self): return self.read_memory_bank()


class raw_write_memory(Dali2Protocol, groups.Base):
    opt_params = {MEMORY_BANK: 0, MEMORY_LOCATION: 8, MEMORY_VALUE: "FF"}
    def exec(self): self.write_memory_bank()


"""https://kiwi.throne.systems/index.php/Tron/Strata/Hardware/Rapida/Resh/Command/Dali#Standard"""


class reset_power_cycle_seen(Dali2Protocol, groups.Standard):
    def exec(self): self._send(device.ResetPowerCycleSeen)


class reset(Dali2Protocol, groups.Standard):
    def exec(self): self._send(device.Reset)


class set_short_address(Dali2Protocol, groups.Standard):
    opt_params = {SHORT_ADDRESS: 'D63'}

    def exec(self):
        try:
            self.set_short_address(self.next_param, not settings.conf.quiescent, not settings.conf.quiescent)
        except errors.AddressAlreadyInUse as ex:
            logger.log.exception(ex)
            swap_short_addresses.help(desc=False, example=False)


class set_operating_mode(Dali2Protocol, groups.Standard):
    opt_params = {OPERATING_MODE: 1}
    def exec(self): self._send(device.SetOperatingMode)


class start_quiescent_mode(Dali2Protocol, groups.Standard):
    def exec(self): self.start_quiescent_mode(self.address)


class stop_quiescent_mode(Dali2Protocol, groups.Standard):
    def exec(self): self.stop_quiescent_mode(self.address)


class enable_power_cycle_notification(Dali2Protocol, groups.Standard):
    def exec(self): self._send(device.EnablePowerCycleNotification)


class disable_power_cycle_notification(Dali2Protocol, groups.Standard):
    def exec(self): self._send(device.DisablePowerCycleNotification)


class save_persistent_variables(Dali2Protocol, groups.Standard):
    def exec(self): self._send(device.SavePersistentVariables)


class query_device_capabilities(Dali2Protocol, groups.Standard):
    def exec(self): return self._send(device_ext.QueryDeviceCapabilities)


class query_device_status(Dali2Protocol, groups.Standard):
    def exec(self): return self._send(device_ext.QueryDeviceStatus)


class query_version_number(Dali2Protocol, groups.Standard):
    def exec(self):
        res = self._send(device.QueryVersionNumber).value.as_integer
        return f'{res >> 2}.{res & 0x03}'


class query_number_of_instances(Dali2Protocol, groups.Standard):
    def exec(self): return self._send(device.QueryNumberOfInstances).value.as_integer


class query_operating_mode(Dali2Protocol, groups.Standard):
    def exec(self): return self._send(device.QueryOperatingMode).value.as_integer


class query_quiescent_mode(Dali2Protocol, groups.Standard):
    def exec(self): return self._send(device_ext.QueryQuiescentMode)


class query_power_cycle_notification(Dali2Protocol, groups.Standard):
    def exec(self): return self._send(device_ext.QueryPowerCycleNotification).value


class query_long_address(Dali2Protocol, groups.Standard):
    def exec(self):
        return {
            'high': self._send(device.QueryRandomAddressH).value.as_integer,
            'middle': self._send(device.QueryRandomAddressM).value.as_integer,
            'low': self._send(device.QueryRandomAddressL).value.as_integer,
        }


class instance_command(Dali2Protocol):
    def_params = DEF_PARAMS_INSTANCE


class enable_instance(instance_command, groups.Standard):
    def exec(self): self._send(device.EnableInstance)


class disable_instance(instance_command, groups.Standard):
    def exec(self): self._send(device.DisableInstance)


class set_primary_instance_group(instance_command, groups.Standard):
    opt_params = {GROUP_ID: '0x31'}
    def exec(self): self._send(device.SetPrimaryInstanceGroup)


class set_instance_group_1(instance_command, groups.Standard):
    opt_params = {GROUP_ID: '0x31'}
    def exec(self): self._send(device.SetInstanceGroup1)


class set_instance_group_2(instance_command, groups.Standard):
    opt_params = {GROUP_ID: '0x31'}
    def exec(self): self._send(device.SetInstanceGroup2)


class set_event_scheme(instance_command, groups.Standard):
    opt_params = {EVENT_SCHEME: '1'}
    def exec(self): self._send(device.SetEventScheme)


class set_event_priority(instance_command, groups.Standard):
    opt_params = {EVENT_PRIORITY: '4'}
    def exec(self): self._send(device.SetEventPriority)


class set_event_filter(instance_command, groups.Standard):
    opt_params = {EVENT_FILTER: '0xFEFEFE'}

    def exec(self):
        try:
            value = int(self.next_param)
        except:
            raise errors.BadParameter()
        self.send(device.DTR2(value >> 16))
        self.send(device.DTR1((value >> 8) & 0x00FF))
        self.send(device.DTR0(value & 0x0000FF))
        self.send(device.SetEventFilter(self.address, self.instance))


class query_instance_type(instance_command, groups.Standard):
    def exec(self): return self.instance_type


class query_instance_error(instance_command, groups.Standard):
    def exec(self): return self._send(device.QueryInstanceError).value.pack


class query_resolution(instance_command, groups.Standard):
    def exec(self): return self._send(device.QueryResolution).value.as_integer


class query_instance_status(instance_command, groups.Standard):
    def exec(self): return self._send(device_ext.QueryInstanceStatus)


class query_instance_enabled(instance_command, groups.Standard):
    def exec(self): return self._send(device_ext.QueryInstanceEnabled).value


class query_primary_instance_group(instance_command, groups.Standard):
    def exec(self): return self._send(device.QueryPrimaryInstanceGroup).value.as_integer


class query_instance_group_1(instance_command, groups.Standard):
    def exec(self): return self._send(device.QueryInstanceGroup1).value.as_integer


class query_instance_group_2(instance_command, groups.Standard):
    def exec(self): return self._send(device.QueryInstanceGroup2).value.as_integer


class query_event_scheme(instance_command, groups.Standard):
    def exec(self): return self._send(device.QueryEventScheme).value.as_integer


class query_event_priority(instance_command, groups.Standard):
    def exec(self): return self._send(device.QueryEventPriority).value.as_integer



"""https://kiwi.throne.systems/index.php/Tron/Strata/Hardware/Rapida/Resh/Command/Dali#Extended"""


class query_dead_time(instance_command, groups.Extended):
    def exec(self):
        self._check_instance(LIGHT, PRESENCE)
        if self.instance_type == LIGHT:
            res = self._send(light_sensor_ext.QueryDeadTime)
        else:
            res = self._send(presence_sensor_ext.QueryDeadTime)
        return res.value.as_integer


class query_report_time(instance_command, groups.Extended):
    def exec(self):
        self._check_instance(LIGHT, PRESENCE)
        if self.instance_type == LIGHT:
            res = self._send(light_sensor_ext.QueryReportTime)
        else:
            res = self._send(presence_sensor_ext.QueryReportTime)
        return res.value.as_integer


class query_hysteresis(instance_command, groups.Extended):
    def exec(self):
        self._check_instance(LIGHT)
        res = self._send(light_sensor_ext.QueryHysteresis)
        return res.value.as_integer


class query_hysteresis_min(instance_command, groups.Extended):
    def exec(self):
        self._check_instance(LIGHT)
        res = self._send(light_sensor_ext.QueryHysteresisMin)
        return res.value.as_integer


class query_hold_time(instance_command, groups.Extended):
    def exec(self):
        self._check_instance(PRESENCE)
        res = self._send(presence_sensor_ext.QueryHoldTime)
        return res.value.as_integer


class query_catching(instance_command, groups.Extended):
    def exec(self):
        self._check_instance(PRESENCE)
        res = self._send(presence_sensor_ext.QueryCatching)
        return res.value.as_integer


class set_hysteresis(instance_command, groups.Extended):
    opt_params = {HYSTERESIS: '1'}

    def exec(self):
        self._check_instance(LIGHT)
        self._send(light_sensor_ext.SetHysteresis)


class set_hysteresis_min(instance_command, groups.Extended):
    opt_params = {HYSTERESIS_MIN: '1'}

    def exec(self):
        self._check_instance(LIGHT)
        self._send(light_sensor_ext.SetHysteresisMin)


class catch_hold_time(instance_command, groups.Extended):
    def exec(self):
        self._check_instance(PRESENCE)
        self._send(presence_sensor_ext.CatchHoldTime)


class set_hold_time(instance_command, groups.Extended):
    opt_params = {HOLD_TIME: '1'}

    def exec(self):
        self._check_instance(PRESENCE)
        self._send(presence_sensor_ext.SetHoldTime)


class cancel_hold_time(instance_command, groups.Extended):
    def exec(self):
        self._check_instance(PRESENCE)
        self._send(presence_sensor_ext.CancelHoldTime)


class set_dead_time(instance_command, groups.Extended):
    opt_params = {DEAD_TIME: '1'}

    def exec(self):
        self._check_instance(LIGHT, PRESENCE)
        if self.instance_type == PRESENCE:
            self._send(presence_sensor_ext.SetDeadTime)
        else:
            self._send(light_sensor_ext.SetDeadTime)


class set_report_time(instance_command, groups.Extended):
    opt_params = {REPORT_TIME: '1'}

    def exec(self):
        self._check_instance(LIGHT, PRESENCE)
        if self.instance_type == PRESENCE:
            self._send(presence_sensor_ext.SetReportTime)
        else:
            self._send(light_sensor_ext.SetReportTime)


class push_buttons_time_getters(instance_command, groups.Extended):
    cmd = None

    def exec(self):
        self._check_instance(BUTTON)
        res = self._send(self.cmd)
        return res.value.as_integer


class query_short_time_min(push_buttons_time_getters):
    cmd = push_button_ext.QueryShortTimeMin


class query_short_time(push_buttons_time_getters):
    cmd = push_button_ext.QueryShortTime


class query_double_time_min(push_buttons_time_getters):
    cmd = push_button_ext.QueryDoubleTimeMin


class query_double_time(push_buttons_time_getters):
    cmd = push_button_ext.QueryDoubleTime


class query_repeat_time(push_buttons_time_getters):
    cmd = push_button_ext.QueryRepeatTime


class query_stuck_time(push_buttons_time_getters):
    cmd = push_button_ext.QueryStuckTime


class _set_push_buttons_time(instance_command, groups.Extended):
    opt_params = {TIME: '0x10'}
    cmd = None

    def exec(self):
        self._check_instance(BUTTON)
        self._send(self.cmd)


class set_short_time(_set_push_buttons_time):
    cmd = push_button_ext.SetShortTime


class set_double_time(_set_push_buttons_time):
    cmd = push_button_ext.SetDoubleTime


class set_repeat_time(_set_push_buttons_time):
    cmd = push_button_ext.SetRepeatTime


class set_stuck_time(_set_push_buttons_time):
    cmd = push_button_ext.SetStuckTime


"""https://kiwi.throne.systems/index.php/Tron/Strata/Hardware/Rapida/Resh/Command/Dali#Wrappers"""


class remove_short_address(Dali2Protocol, groups.Wrappers):
    def exec(self):
        self.send(device.DTR0(0xff))
        self.send(device.SetShortAddress(self.address))


class add_to_group(Dali2Protocol, groups.Wrappers):
    opt_params = {GROUP_IDs: '0 2 31'}

    def exec(self):
        _dtr_0_15 = 0x0
        _dtr_16_31 = 0x0

        err = []

        for group_id in self.params:
            if not str(group_id).isdigit() or int(group_id) not in range(0, 32):
                err.append(group_id)
            group_id = int(group_id)

            if group_id < 16:
                _dtr_0_15 += 1 << group_id
            else:
                _dtr_16_31 += 1 << (group_id % 16)

        if len(err) > 0:
            raise errors.BadParameter('GroupIds required as [<int>0..31]. Received {}'.format(err))

        if _dtr_0_15 > 0:
            self._params = list(_dtr_0_15.to_bytes(2, 'big'))
            self._send(device.AddToDeviceGroupsZeroToFifteen)
        if _dtr_16_31 > 0:
            self._params = list(_dtr_16_31.to_bytes(2, 'big'))
            self._send(device.AddToDeviceGroupsSixteenToThirtyOne)


class remove_from_group(Dali2Protocol, groups.Wrappers):
    opt_params = {GROUP_IDs: '0 2 31'}

    def exec(self):
        _dtr_0_15 = 0x0
        _dtr_16_31 = 0x0

        err = []

        for group_id in self.params:
            if not str(group_id).isdigit() or int(group_id) not in range(0, 32):
                err.append(group_id)
            group_id = int(group_id)

            if group_id < 16:
                _dtr_0_15 += 1 << group_id
            else:
                _dtr_16_31 += 1 << (group_id % 16)

        if len(err) > 0:
            raise errors.BadParameter('GroupIds required as [<int>0..31]. Received {}'.format(err))

        if _dtr_0_15 > 0:
            self._params = list(_dtr_0_15.to_bytes(2, 'big'))
            self._send(device.RemoveFromDeviceGroupsZeroToFifteen)
        if _dtr_16_31 > 0:
            self._params = list(_dtr_16_31.to_bytes(2, 'big'))
            self._send(device.RemoveFromDeviceGroupsSixteenToThirtyOne)


class query_device_groups(Dali2Protocol, groups.Wrappers):
    def exec(self):
        result = tools.bits_indexes(
            self._send(device.QueryDeviceGroupsZeroToSeven)
                .value.as_integer, start_from=0
        ) + tools.bits_indexes(
            self._send(device.QueryDeviceGroupsEightToFifteen)
                .value.as_integer, start_from=8
        ) + tools.bits_indexes(
            self._send(device.QueryDeviceGroupsSixteenToTwentyThree)
                .value.as_integer, start_from=16
        ) + tools.bits_indexes(
            self._send(device.QueryDeviceGroupsTwentyFourToThirtyOne)
                .value.as_integer, start_from=24
        )
        return result


class query_feature_types(Dali2Protocol, groups.Wrappers):
    def exec(self):
        res = []
        resp = self._send(device.QueryFeatureType)
        if resp.value.as_integer == 0xff:
            _r = None
            while _r is None or _r.value.as_integer == 0xfe or _r.value.as_integer == 0xff:
                _r = self._send(device.QueryNextFeatureType)
                if _r is None or _r.value is None or _r.value.as_integer == 0xfe:
                    break
                res.append(_r.value.as_integer)
        elif resp.value.as_integer != 0xfe:
            _arr = [resp.value.as_integer]

        return res


class query_instance_input(instance_command, groups.Wrappers):
    def exec(self):
        resolution = query_resolution(*self._args).exec()
        r = self._send(device.QueryInputValue)
        _val = bin(r.value.as_integer)[2:].rjust(8, '0')
        while True:
            try:
                r = self._send(device.QueryInputValueLatch)
            except errors.DeviceNotAnswered:
                break
            _val += bin(r.value.as_integer)[2:].rjust(8, '0')

        _val = _val.ljust(resolution, '0')[:resolution]

        return {v.EVENT: int(_val[:resolution], 2), v.FULL: int(_val, 2), v.MASK: _val}


class query_event_filter(instance_command, groups.Wrappers):
    def exec(self):
        res = self._send(device.QueryEventFilterZeroToSeven).value.as_integer
        try:
            res += self._send(device.QueryEventFilterEightToFifteen).value.as_integer << 8
            res += self._send(device.QueryEventFilterSixteenToTwentyThree).value.as_integer << 16
        finally:
            return res


class read_firmware_version(Dali2Protocol, groups.Wrappers):
    def exec(self):
        result = b''
        for addr in range(0x9, 0xb):
            result += self.read_memory_bank(0, addr)
        return '.'.join(str(b) for b in result)


class read_hardware_version(Dali2Protocol, groups.Wrappers):
    def exec(self):
        result = b''
        for addr in range(0x13, 0x15):
            result += self.read_memory_bank(0, addr)
        return '.'.join(str(b) for b in result)


class read_gtin(Dali2Protocol, groups.Wrappers):
    def exec(self):
        result = b''
        for addr in range(0x03, 0x09):
            result += self.read_memory_bank(0, addr)
        return ''.join(hex(b)[2:].rjust(2, '0') for b in result).upper()


class read_gtin_oem(Dali2Protocol, groups.Wrappers):
    def exec(self):
        result = b''
        for addr in range(0x03, 0x09):
            result += self.read_memory_bank(1, addr)
        return ''.join(hex(b)[2:].rjust(2, '0') for b in result).upper()


class read_serial(Dali2Protocol, groups.Wrappers):
    def exec(self):
        result = b''
        for addr in range(0x0b, 0x13):
            result += self.read_memory_bank(0, addr)
        return ''.join(hex(b)[2:].rjust(2, '0') for b in result).upper()


class read_serial_oem(Dali2Protocol, groups.Wrappers):
    def exec(self):
        result = b''
        for addr in range(0x9, 0x11):
            result += self.read_memory_bank(1, addr)
        return ''.join(hex(b)[2:].rjust(2, '0') for b in result).upper()


"""https://kiwi.throne.systems/index.php/Tron/Strata/Hardware/Rapida/Resh/Command/Dali2#Wizards"""


class identify(Dali2Protocol, groups.Wizards):
    var_params = {TIMEOUT: 9}

    def __init__(self, *args, **kwargs):
        super(identify, self).__init__(*args, **kwargs)
        self.__stopped = False
        self.__repeater = None

    """Не доступно для групп"""
    def exec(self):
        if isinstance(self.address, dali.address.Group):
            raise errors.AddressAlreadyInUse(self.address)

        _broadcast = isinstance(self.address, dali.address.Broadcast)
        _address = None
        if not _broadcast:
            _address = self.address.address

        if len(self.params) > 0:
            timeout = self.next_param
            if not isinstance(timeout, int):
                raise errors.BadParameter()
        else:
            timeout = 3600

        if timeout > 0:
            self.__start()

            if threading.current_thread() is threading.main_thread():
                logger.log.question(loc.IDENTIFY_QUESTION, timeout)
            else:
                sleep(timeout)
            self.stop()
        else:
            self.send(device.IdentifyDevice(self.address))

    def stop(self):
        self.__stopped = True
        if self.__repeater:
            self.__repeater.cancel()
            self.__repeater.join()
            self.__repeater = None
        self.send(device.Terminate())

    def __start(self):
        if self.__stopped is False:
            self.send(device.IdentifyDevice(self.address))
            self.__repeater = threading.Timer(8, function=self.__start)
            self.__repeater.start()
        else:
            self.stop()


class swap_short_addresses(Dali2Protocol, groups.Wizards):
    def_params = {MODULE: 1, CHANNEL: 2}
    opt_params = {ADDRESS1: 'D0', ADDRESS2: 'D63'}

    def exec(self):
        self._address = tools.parse_address(self.next_param)
        raw_new_address = self.next_param
        new_short = tools.parse_address(raw_new_address)
        if not isinstance(new_short, dali.address.Short):
            raise errors.BadParameter()

        try:
            self.set_short_address(raw_new_address, False)
        except errors.AddressAlreadyInUse:
            old_short = self.address

            old_long = []
            new_long = []

            self.send(device.Terminate())

            self.send(device.Initialise(old_short.address))
            self.send(device.Initialise(new_short.address))

            old_long.append(self.send(device.QueryRandomAddressH(old_short)))
            old_long.append(self.send(device.QueryRandomAddressM(old_short)))
            old_long.append(self.send(device.QueryRandomAddressL(old_short)))

            new_long.append(self.send(device.QueryRandomAddressH(new_short)))
            new_long.append(self.send(device.QueryRandomAddressM(new_short)))
            new_long.append(self.send(device.QueryRandomAddressL(new_short)))

            self.send(device.SearchAddrH(old_long[0].value.as_integer))
            self.send(device.SearchAddrM(old_long[1].value.as_integer))
            self.send(device.SearchAddrL(old_long[2].value.as_integer))
            self.send(device.ProgramShortAddress(new_short.address))

            self.send(device.SearchAddrH(new_long[0].value.as_integer))
            self.send(device.SearchAddrM(new_long[1].value.as_integer))
            self.send(device.SearchAddrL(new_long[2].value.as_integer))
            self.send(device.ProgramShortAddress(old_short.address))

            self.send(device.Terminate())


class line_commands(Dali2Protocol, groups.Wizards):
    def_params = {MODULE: 1, CHANNEL: 2}
    opt_params = {MODE: MODE_CROSS}
    var_params = {START_ADDRESS: 0}
    __bus = None
    __is_cross_addressing = None
    __start_index = None

    @property
    def _bus(self):
        if self.__bus is None:
            self.__bus = Bus2(self, self._is_cross_addressing, self._start_index, step_handler=self.on_step)
        return self.__bus

    @property
    def _is_cross_addressing(self):
        if self.__is_cross_addressing is None:
            cross_addr = str(self.next_param).upper()
            if cross_addr == MODE_NOCROSS:
                self.__is_cross_addressing = False
            elif cross_addr == MODE_CROSS:
                self.__is_cross_addressing = True
            else:
                raise errors.BadParameter()

        return self.__is_cross_addressing

    @property
    def _start_index(self):
        if self.__start_index is None:
            if len(self.params) > 0:
                index = self.next_param
                if isinstance(index, int):
                    self.__start_index = index
                else:
                    raise errors.BadParameter()
            else:
                self.__start_index = 0

        return self.__start_index


class scan_line(line_commands):
    opt_params = {}
    var_params = {}

    def exec(self):
        scan = []
        for addr in range(0, 64):
            self._address = dali.address.Short(addr)
            try:
                self._instance_count = None
                count = self.instance_count
            except:
                continue
            else:
                dev = {'address': addr, 'types': []}
                for index in range(0, count):
                    try:
                        self._instance_type = None
                        self._instance = dali.address.InstanceNumber(index)
                        dev['types'].append(self.instance_type)
                    except:
                        continue

                scan.append(dev)

        return scan

    @property
    def _is_cross_addressing(self): return False

    @property
    def _start_index(self): return 0


class test_line(Dali2Protocol, groups.Wizards):
    DEF_COUNT = 10
    DEF_COMMAND = device.QueryNumberOfInstances

    def_params = {MODULE: 1, CHANNEL: 2}
    var_params = {COUNT: DEF_COUNT, COMMAND: DEF_COMMAND.__name__, ADDRESSES: 'D0,D64', SORT_BY_ADDR: ''}

    def _step_handler(self, progress, *args, **kwargs):
        if progress < 100:
            sys.stdout.write('\rProgress: {}%'.format(progress))
        else:
            sys.stdout.write('\r' + ' ' * 14 + '\r')

        sys.stdout.flush()

    def exec(self):
        sort_by_addr = False
        if SORT_BY_ADDR in self._args:
            sort_by_addr = True
            self._args.remove(SORT_BY_ADDR)

        count = self.DEF_COUNT
        command = self.DEF_COMMAND

        if self.has_next:
            if isinstance(self.next_param_recursive, int):
                count = self.next_param
                if count < 1:
                    raise errors.BadParameter('Количество запросов должно быть больше 0')

        if self.has_next:
            commands = {
                cmd.__name__: cmd
                for cmd in device._StandardDeviceCommand.__subclasses__()
                if issubclass(cmd, device._StandardDeviceCommand) and cmd._response is not None
            }

            if self.next_param_recursive in commands:
                command = commands[self.next_param]

        if self.has_next:
            try:
                addresses = sorted(
                    [
                        tools.parse_address(a.strip())
                        for a in self.next_param.split(',')
                    ],
                    key=lambda a: a.address
                )

            except:
                raise errors.BadParameter('Ошибка парсинга адресов')
        else:
            addresses = [dali.address.Short(i) for i in range(0, 64)]

        result = {}
        step = 0
        self.start_quiescent_mode()
        for i in range(0, count):
            if self.has_break:
                break
            for addr in addresses:
                if self.has_break:
                    break

                step += 1

                result.setdefault(addr.address, 0)
                try:
                    self._address = addr
                    self._send(command)
                    result[addr.address] += 1
                except BaseException:
                    continue
                finally:
                    if not logger.log.is_json:
                        self.on_step(min(100, max(1, int(step / (count * len(addresses)) * 100))))
        self.stop_quiescent_mode()

        result = {a: '{}%'.format(int(v * 100 / count)) for a, v in result.items()}

        if not sort_by_addr:
            result = dict(sorted(result.items(), key=lambda item: item[1], reverse=True))

        return result


class construct_line(line_commands):
    def exec(self): return self._bus.assign_short_addresses(False, address=0xff)


class extend_line(line_commands):
    def exec(self): return self._bus.assign_short_addresses(True, address=0xff >> 1)


class fix_double_addresses(line_commands):
    def exec(self): return self._bus.address_resolution(address=0xff)


class dump(Dali2Protocol, groups.Wizards):
    @staticmethod
    def queries_device(): return {
        v.SERIAL: dict(get=read_serial),
        v.GTIN: dict(get=read_gtin),
        v.HARDWARE: dict(get=read_hardware_version),
        v.FIRMWARE: dict(get=read_firmware_version),
        v.GTIN_OEM: dict(get=read_gtin_oem),
        v.SERIAL_OEM: dict(get=read_serial_oem),
        v.OPERATING_MODE: dict(get=query_operating_mode, set=set_operating_mode),
        v.GROUPS: dict(get=query_device_groups),
        v.IDENTIFY: dict(set=identify),
    }

    @staticmethod
    def queries_instance(): return {
        v.IS_ENABLED: dict(get=query_instance_enabled),
        v.EVENT_SCHEME: dict(get=query_event_scheme, set=set_event_scheme),
        v.EVENT_PRIORITY: dict(get=query_event_priority, set=set_event_priority),
        v.EVENT_FILTER: dict(get=query_event_filter, set=set_event_filter),
        v.INSTANCE_GROUP_0: dict(get=query_primary_instance_group, set=set_primary_instance_group),
        v.INSTANCE_GROUP_1: dict(get=query_instance_group_1, set=set_instance_group_1),
        v.INSTANCE_GROUP_2: dict(get=query_instance_group_2, set=set_instance_group_2),
        v.RESOLUTION: dict(get=query_resolution),
        v.FEATURE_TYPES: dict(get=query_feature_types),
    }

    @staticmethod
    def queries_instance_type(): return {
        LIGHT: {
            v.DEAD_TIME: dict(get=query_dead_time, set=set_dead_time),
            v.REPORT_TIME: dict(get=query_report_time, set=set_report_time),
            v.HYSTERESIS: dict(get=query_hysteresis, set=set_hysteresis),
            v.HYSTERESIS_MIN: dict(get=query_hysteresis_min, set=set_hysteresis_min),
        },
        PRESENCE: {
            v.DEAD_TIME: dict(get=query_dead_time, set=set_dead_time),
            v.REPORT_TIME: dict(get=query_report_time, set=set_report_time),
            v.CATCHING: dict(get=query_catching),
            v.HOLD_TIME: dict(get=query_hold_time, set=set_hold_time),
        },
        BUTTON: {
            v.SHORT_TIME_MIN: dict(get=query_short_time_min),
            v.DOUBLE_TIME_MIN: dict(get=query_double_time_min),
            v.SHORT_TIME: dict(get=query_short_time, set=set_short_time),
            v.DOUBLE_TIME: dict(get=query_double_time, set=set_double_time),
            v.REPEAT_TIME: dict(get=query_repeat_time, set=set_repeat_time),
            v.STUCK_TIME: dict(get=query_stuck_time, set=set_stuck_time),
        }
    }

    def __init__(self, *args, **kwargs):
        super(dump, self).__init__(*args, **kwargs)
        self._custom_stack = {}

    @property
    def custom_stack(self): return self._custom_stack

    @custom_stack.setter
    def custom_stack(self, stack): self._custom_stack = stack

    def need_key(self, key):
        return not self._custom_stack or key in self._custom_stack

    def exec(self):
        root_address = self.address
        devices = []
        to_rise = None

        if self.address.__class__ == dali.address.Broadcast:
            devices = [dali.address.Short(i) for i in range(0, 64)]
        elif self.address.__class__ in (dali.address.Group, dali.address.BroadcastUnaddressed):
            raise errors.CustomError('Can not fill dump for {}'.format(self.address.__class__.__name__))
        else:
            devices.append(self.address)

        result = dict()

        self.start_quiescent_mode()

        for addr in devices:
            if self.has_break:
                break

            self._address = addr
            try:
                count = self.instance_count
                self.on_step(addr.address, v.NUMBER_OF_INSTANCES, count)
            except errors.DeviceNotAnswered as ex:
                if isinstance(root_address, dali.address.Short):
                    to_rise = ex
                    break
            else:
                result[addr.address] = {}
                self.__on_progress(f'{addr.address}:')

                for key, query in self.queries_device().items():
                    if self.has_break:
                        break

                    if 'get' in query and self.need_key(key):
                        try:
                            val = self.__exec_query(query['get'], addr)
                        except BaseException as ex:
                            self.on_step(addr.address, key, ex)
                            self.__on_progress(f'  {key}: {ex}')
                            val = str(ex)
                        else:
                            self.on_step(addr.address, key, val)
                            self.__on_progress(f'  {key}: {val}')

                        result[addr.address][key] = val

                self.__on_progress(f'  instances:')
                result[addr.address]['instances'] = []
                _instance = {}
                for i in range(0, count):
                    self.__on_progress(f'  index: {i}')
                    queries_instance = self.queries_instance()
                    try:
                        _instance_type = self.__exec_query(query_instance_type, addr, i)
                    except BaseException as ex:
                        self.__on_progress(f'    type: {ex}')
                    else:
                        self.__on_progress(f'    type: {_instance_type}')
                        self.on_step(addr.address, 'type', _instance_type, index=i)
                        queries_instance.update(self.queries_instance_type().get(_instance_type, {}))

                        for key, query in queries_instance.items():
                            if 'get' in query and self.need_key(key):
                                try:
                                    val = self.__exec_query(query['get'], addr, i)
                                except BaseException as ex:
                                    self.on_step(addr.address, key, ex, index=i)
                                    self.__on_progress(f'    {key}: {str(ex)}')
                                    val = str(ex)
                                else:
                                    self.on_step(addr.address, key, val, index=i)
                                    self.__on_progress(f'    {key}: {val}')

                                _instance[key] = val

                    result[addr.address]['instances'].append(_instance)

        self.stop_quiescent_mode()

        if isinstance(to_rise, BaseException):
            raise to_rise
        elif logger.log.is_json:
            return result

    def __exec_query(self, query, addr, index=None):
        query = query()
        if isinstance(query, Dali2Protocol):
            query._module = self.module
            query._channel = self.channel
            query._address = addr
            if index is not None:
                query._instance = dali.address.InstanceNumber(index)
            return query.exec()

    def __on_progress(self, string):
        if not logger.log.is_json:
            print(string)
