import ipaddress
import os

from resh.domains._domain import Domain
from resh.utils import groups, compat, settings, errors
from resh.utils.const import SYSTEM, IP, DAY, YEAR, MONTH, MINUTES, HOURS, SECONDS, GATEWAY_VAR, \
    MASK_LC, DNS, DNS2, TZ

DHCPCD = '/etc/dhcpcd.conf'


@compat.add_metaclass(compat.DomainTracker)
class SystemDomain(Domain, groups.System):
    domain_name = SYSTEM

    TYPE_DEST = '/boot/rapida/'
    TYPE_MASK = 'device-type-'


""" https://kiwi.throne.systems/index.php/Tron/Strata/Hardware/Rapida/Resh/Command/System#System """


class get_device_type(SystemDomain):

    def exec(self):
        for f in os.popen('sudo ls %s | grep "%s"' % (self.TYPE_DEST, self.TYPE_MASK)).read().strip().split():
            try:
                return int(f.split('device-type-')[1])
            except: pass

        return 1


class get_mac(SystemDomain):
    def exec(self): return os.popen('sudo ifconfig -a | grep ether').read().strip().split()[1].upper()


class get_ip(SystemDomain):
    def exec(self):
        res = {}
        try:
            res[IP], mask = os.popen('grep -i ^"static ip_address" %s' % DHCPCD).read().strip().split('=')[1].split('/')
            mask = ('1' * int(mask)).ljust(32, '0')
            res[MASK_LC] = '.'.join([f'{int(mask[i * 8:(i + 1) * 8], 2)}' for i in range(0, 4)])
            res[GATEWAY_VAR] = os.popen('grep -i ^"static routers=" %s' % DHCPCD).read().strip().split('=')[1]
        finally:
            return res


class set_ip(SystemDomain):
    opt_params = {IP: '192.168.1.190'}
    var_params = {MASK_LC: '255.255.225.0', GATEWAY_VAR: '192.168.1.1'}

    def exec(self):
        try:
            ip = str(ipaddress.ip_address(self.next_param))
            mask = ''
            gateway = ''
            if self.has_next:
                mask = str(ipaddress.ip_address(self.next_param))
                mask = ''.join(bin(int(b))[2:].rjust(8, '0') for b in mask.split('.')) + '0'
                mask = len(mask[:mask.index('0')])
                if self.has_next:
                    gateway = str(ipaddress.ip_address(self.next_param))
        except:
            raise errors.BadParameter()

        last_ip = os.popen(f'cat {DHCPCD} | grep "^static ip_address="').read().strip()
        if not mask:
            mask = last_ip.split('/')[1]
        new_ip = f'static ip_address={ip}/{mask}'
        r = os.popen(f'sudo sed -i -e "s@^{last_ip}\\b@{new_ip}@g" {DHCPCD}').read().strip()
        if r: raise errors.CustomError(r)

        if gateway:
            last_gw = os.popen(f'cat {DHCPCD} | grep "^static routers="').read().strip()
            new_gw = f'static routers={gateway}'
            r = os.popen(f'sudo sed -i -e "s@^{last_gw}\\b@{new_gw}@g" {DHCPCD}').read().strip()
            if r: raise errors.CustomError(r)

        os.system('sudo ip addr flush dev eth0 && sudo systemctl restart dhcpcd.service')


class get_dns(SystemDomain):
    def exec(self):
        return os.popen('grep -i ^"static domain_name_servers=" %s' % DHCPCD).read().strip().split('=')[1].split()


class set_dns(SystemDomain):
    opt_params = {DNS: '4.4.4.4'}
    var_params = {DNS2: '8.8.8.8'}

    def exec(self):
        arr = []
        while self.has_next:
            try: arr.append(str(ipaddress.ip_address(self.next_param)))
            except ValueError: raise errors.BadParameter()

        last_dns = os.popen(f'cat {DHCPCD} | grep "^static domain_name_servers="').read().strip()
        new_dns = f'static domain_name_servers={" ".join(arr)}'
        r = os.popen(f'sudo sed -i -e "s@^{last_dns}\\b@{new_dns}@g" {DHCPCD}').read().strip()
        if r:
            raise errors.CustomError(r)


class get_date_time(SystemDomain):
    def exec(self):
        tz = os.popen('cat /etc/timezone').read().strip()
        return dict(datetime=os.popen(f'TZ={tz} date --iso-8601=seconds').read().strip(), timezone=tz)


class set_date(SystemDomain):
    opt_params = {DAY: 31, MONTH: 12}
    var_params = {YEAR: 1970}

    def exec(self):
        try:
            day = str(self.next_param).rjust(2, '0')
            month = str(self.next_param).rjust(2, '0')
            try:
                year = str(self.next_param)
            except errors.BadParameter:
                year = os.popen('date +%Y').read().strip()
        except:
            raise errors.BadParameter()

        tz = os.popen('cat /etc/timezone').read().strip()
        time = os.popen(f'TZ={tz} date +%H:%M:%S').read().strip()
        date = f'{year}-{month}-{day}'

        # os.popen('sudo timedatectl set-ntp off')
        res = os.popen(f'TZ={tz} date +%D -s "{time} {date}"').read().strip()
        if res and res != '{}/{}/{}'.format(month, day, year[-2:]):
            raise errors.BadParameter(res)

        os.popen(f'sudo hwclock --systohc')
        # os.popen('sudo sudo timedatectl set-ntp on')


class set_time(SystemDomain):
    opt_params = {HOURS: 23, MINUTES: 59}
    var_params = {SECONDS: 59}

    def exec(self):
        time = ':'.join([str(c).rjust(2, '0') for c in self.params])

        # os.popen('sudo timedatectl set-ntp off')
        tz = os.popen('cat /etc/timezone').read().strip()
        res = os.popen(f'TZ={tz} date +%T -s "{time}"').read().strip()
        if res and res[:len(time)] != time:
            raise errors.BadParameter()

        os.popen(f'sudo hwclock --systohc')
        # os.popen('sudo sudo timedatectl set-ntp on')


class set_time_zone(SystemDomain):
    opt_params = {TZ: '+3'}

    @staticmethod
    def __do_raise(): raise errors.BadParameter()

    def exec(self):
        try:
            tz = int(self.next_param)
            tz = f'GMT{-tz}' if tz > 0 else f'GMT+{-tz}'

            if os.path.exists(f'/usr/share/zoneinfo/Etc/{tz}'):
                os.system(f'sudo timedatectl set-timezone Etc/{tz}')
            else: raise
        except:
            raise errors.BadParameter()


class get_info(SystemDomain):
    def exec(self):
        return dict(model=os.popen('grep -i "Product: " /boot/rapida/device-info').read().strip().split('Product: ')[1],
                    imageVersion=os.popen('cat /boot/rapida/image-issue').read().strip())


class get_status(SystemDomain):
    def exec(self):
        res = {}
        try:
            flash = os.popen('df -h --output=source,size,used,avail,pcent | grep /dev/root').read().strip()
            while '  ' in flash:
                flash = flash.replace('  ', ' ')
            flash = flash.split('/dev/root')[1].strip().split(' ')

            res['flash'] = dict(total=flash[0], used=flash[1], free=flash[2])
        except:
            pass

        try:
            res['temp'] = os.popen('/opt/vc/bin/vcgencmd measure_temp').read().strip().split('=')[1]
        except:
            pass
        try:
            res['can'] = 'OK' if os.popen(f'sudo ifconfig {settings.conf.interface} | grep UP,RUNNING').read().strip() else 'not present'
        except:
            pass

        try:
            res['sd'] = 'OK ({}G)'.format(round(int(os.popen('sudo blockdev --getsize64 /dev/mmcblk1').read().strip())/(1024**3), 1))
        except:
            res['sd'] = 'not present'

        return res


class reboot(SystemDomain):
    def exec(self): os.system('sudo reboot')

