import threading

import can

from resh.utils import const
from resh.protocols.can_utils import CONTROLLER, TYPES
from resh.utils.errors import DeviceNotAnswered, CanNotCommitted, CanOpenError
from resh.utils.logger import log
from resh.utils.settings import conf
from resh.utils.tools import bytes2str


class CanPacket:
    CONTROLLER_ID = 31
    pr = 0

    def __init__(self, from_id, to_id, module_type, end_frame, cmd):
        self.from_id = from_id
        self.to_id = to_id
        self.module_type = module_type
        self.end_frame = end_frame
        self.cmd = cmd

    def __str__(self):
        return bytes2str(self.frame)

    def __repr__(self):
        return f'[{self.__class__.__name__}]: {self.__str__()}'

    @property
    def module_address(self):
        if self.from_id == CONTROLLER:
            return self.to_id
        else:
            return self.from_id

    @property
    def channel(self):
        return None

    @property
    def arbitration_id(self):
        return (self.pr << 10) + (self.from_id << 5) + self.to_id

    @property
    def module_desc(self):
        tp = self.module_type
        if tp in TYPES:
            return TYPES[tp]['desc']
        return f'Unknown module type "{hex(tp)}"'

    @property
    def module_name(self):
        tp = self.module_type
        if tp in TYPES:
            return TYPES[tp]['name']
        return f'Unknown module type "{hex(tp)}"'

    @property
    def frame(self) -> bytearray:
        raise NotImplementedError()

    @property
    def data(self) -> bytearray:
        raise NotImplementedError()

    @property
    def payload(self):# -> Payload:
        raise NotImplementedError()

    @staticmethod
    def str_can_message(msg):
        return ('{}  {}   [{}]  {}'.format(conf.interface,
                                           hex(msg.arbitration_id)[2:].rjust(3, '0').upper(),
                                           len(msg.data),
                                           ' '.join(hex(b)[2:].rjust(2, '0').upper() for b in msg.data)).upper())


class CanSend(CanPacket):
    commit_timeout = 2
    answer_timeout = 0.5

    @staticmethod
    def filters(arbitration_id: int) -> list:
        return [{"can_id": arbitration_id, "can_mask": 1023}]

    def __init__(self, module_type, module_address, end_frame, cmd=0, **kwargs):
        super(CanSend, self).__init__(CONTROLLER, module_address, module_type, end_frame, cmd)
        self.bus_channel = conf.interface
        self.can_error = None
        self.answer_timer = None
        self.commit_timer = None
        self.need_state = 1
        self.need_answer = False
        self.listen_service = False
        self.state = 0
        self.stack = None
        self._bus = None
        self.send_twice = False

        for key, value in kwargs.items():
            self.__setattr__(key, value)

    def start_answer_timer(self):
        if self.answer_timer:
            self.answer_timer.cancel()

        self.answer_timer = threading.Timer(self.answer_timeout, function=self.on_timeout, args=[DeviceNotAnswered])
        self.answer_timer.start()

    def stop_answer_timer(self):
        if self.answer_timer:
            self.answer_timer.cancel()
        self.answer_timer = None

    def start_commit_timer(self):
        self.stop_commit_timer()
        self.commit_timer = threading.Timer(self.commit_timeout, function=self.on_timeout, args=[CanNotCommitted])
        self.commit_timer.start()

    def stop_commit_timer(self):
        if self.commit_timer:
            self.commit_timer.cancel()
        self.commit_timer = None

    @property
    def _bus_filters(self) -> list:
        return [
            {
                "can_id": (self.module_address << 5) + self.CONTROLLER_ID,
                "can_mask": 1023,
            },
        ]

    @property
    def bus(self):
        if self._bus is None:
            self._bus = can.interface.Bus(
                channel=self.bus_channel,
                bustype=const.BUS_TYPE,
                can_filters=self._bus_filters
            )

        return self._bus

    def on_timeout(self, err):
        self.can_error = err

    def check_commit(self, dump):
        raise NotImplementedError()

    @property
    def committed(self):
        return self.state == self.need_state

    def on_response(self, frame):
        raise NotImplementedError()

    def collect(self, a_id, dump):
        if self.stack is None:
            self.stack = dict()
        mid = (a_id >> 5) & 0x1F
        if mid not in self.stack:
            self.stack[mid] = []
        self.stack[mid].append(dump)

    def receive(self):
        while self.can_error is None:
            resp = self.bus.recv(0.05)
            if resp is not None:
                log.can(self.str_can_message(resp))
                dump = CanDump.can_dump(resp)
                if self.to_id == 0:
                    self.collect(resp.arbitration_id, dump)
                elif self.to_id == dump.from_id:
                    if not self.committed:
                        self.check_commit(dump)
                        if self.committed:
                            if not self.need_answer:
                                return
                            else:
                                self.start_answer_timer()
                    else:
                        result = self.on_response(dump)
                        if result is not None:
                            return result

        raise self.can_error(stack=self.stack)

    def send(self, need_answer, listen_service=False):
        self.need_answer = need_answer
        self.listen_service = listen_service
        self.need_state = 2 if self.send_twice else self.need_state

        try:
            message = can.Message(arbitration_id=self.arbitration_id,
                                  data=self.frame,
                                  extended_id=False)

            self.bus.send(message)
            log.can(self.str_can_message(message))

            if self.committed:
                self.start_answer_timer()
            else:
                self.start_commit_timer()

            if not self.committed or self.need_answer:
                return self.receive()
        except OSError as ex:
            raise CanOpenError(reason=str(ex))

    def send_raw(self):
        try:
            message = can.Message(arbitration_id=self.arbitration_id,
                                  data=self.end_frame,
                                  extended_id=False)
            self.bus.send(message)
            log.can(self.str_can_message(message))
        except OSError as ex:
            raise CanOpenError(reason=str(ex))


class CanDump(CanPacket):
    def __init__(self, from_id, to_id, module_type, frame, cmd=0):
        super(CanDump, self).__init__(from_id, to_id, module_type, frame, cmd)

    @property
    def payload(self):
        raise NotImplementedError()

    @property
    def frame(self):
        return self.end_frame

    @staticmethod
    def can_dump(message: can.Message) -> CanPacket:
        from_id = (message.arbitration_id >> 5) & CanPacket.CONTROLLER_ID
        to_id = message.arbitration_id & CanPacket.CONTROLLER_ID
        frame = message.data
        b0 = frame[0]
        cmd = b0 >> 7
        cl = b0 & 31

        if cmd == 0:
            if cl == 1:
                if from_id == CONTROLLER:
                    from .dali_packet import DaliSend
                    channel = DaliSend.channel_from_frame(frame)
                    return DaliSend(cl, to_id, channel, frame[3:], need_answer=DaliSend.need_answer_flag_from_frame(frame))
                else:
                    """ CMD = 0 """
                    from .dali_packet import DaliDump
                    return DaliDump(from_id, to_id, cl, frame, cmd)
            elif cl == 2:
                if from_id == CONTROLLER:
                    from .modbusi_packet import ModbusSend
                    channel = ModbusSend.channel_from_frame(frame)
                    return ModbusSend(to_id, channel, frame[2:])
                else:
                    """ CMD = 0 """
                    from .modbusi_packet import ModbusDump
                    return ModbusDump(from_id, to_id, cl, frame, cmd)
        else:
            """ CMD = 1"""
            from .rapida_packet import RapidaDump
            return RapidaDump(from_id, to_id, cl, frame, cmd)
