from typing import Union

from resh.protocols.can_utils.rapida.responses import ECService, EEService
from resh.protocols.payloads.rapida import ServicePacket
from resh.protocols.can_utils.rapida_packet import RapidaDump
import resh.protocols.dali_utils.external_commands as ext
from resh.protocols.payloads import Payload
from resh.utils.errors import DeviceNotAnswered, CanError, CanCollision
from resh.utils.logger import log
from .can_packet import CanSend, CanDump
from ..payloads import dali as dali_payload


class DaliSend(CanSend):
    def __init__(self, module_type, module_address, channel, end_frame, need_answer=False):
        super(DaliSend, self).__init__(module_type, module_address, end_frame, cmd=0)
        self.need_answer = need_answer
        self._channel = channel

    @property
    def is_dali2(self):
        return len(self.end_frame) == 3

    def check_commit(self, dump):
        if isinstance(dump, DaliDump) and dump.channel == self.channel:
            if dump.data == self.end_frame:
                self.state += 1
                log.info(f'  <- ✓: {" ".join([hex(b)[2:].rjust(2, "0") for b in dump.data])}')
                if self.committed:
                    self.stop_commit_timer()

    def on_response(self, dump):
        if isinstance(dump, DaliDump) and dump.channel == self.channel:
            if dump.is_dali2():
                return
            elif not self.listen_service and dump.is_error:
                raise CanError()
            elif len(dump.data) == 0:
                raise DeviceNotAnswered()
            elif len(dump.data) == 1:
                self.stop_answer_timer()
                return dump.data
        elif self.listen_service and isinstance(dump, RapidaDump) and isinstance(dump.payload, ServicePacket):
            if isinstance(dump.payload.command, EEService):
                raise CanError()
            elif isinstance(dump.payload.command, ECService):
                raise CanCollision()

    @property
    def frame(self):
        b1 = 0x0
        b2 = 0x1 << self.channel - 1

        if self.send_twice:
            b1 = b1 | 0x1
        if self.is_dali2:
            b1 = b1 | 0x2
        if self.need_answer:
            b1 = b1 | 0x4

        frame = bytearray()
        frame.append(self.module_type)
        frame.append(b1)
        frame.append(b2)
        return frame + self.end_frame

    @property
    def payload(self):
        return generate_payload(self)

    @property
    def channel(self):
        return self._channel

    @property
    def data(self) -> bytearray:
        return self.end_frame

    @staticmethod
    def channel_from_frame(frame: bytearray) -> int:
        return 8 - bin(frame[2])[2:].rjust(8, '0').rindex('1')

    @staticmethod
    def need_answer_flag_from_frame(frame: bytearray) -> int:
        return frame[1] & 0x04 == 0x04


class DaliDump(CanDump):

    def is_dali2(self):
        return (self.end_frame[1] >> 3) == 0x3

    @property
    def channel(self):
        return (self.frame[1] & 0x7) + 1

    @property
    def data(self):
        size_byte = self.frame[1] & 24
        if self.is_error:
            return b''
        elif size_byte == 0:
            size = 1
        elif size_byte == 8:
            size = 2
        elif size_byte == 24:
            size = 3
        else:
            size = 0
        return self.frame[2:2+size]

    @property
    def is_error(self): return self.frame[1] & 32 == 32

    @property
    def payload(self): return generate_payload(self)


last_queries = {}
device_type = {}


def generate_payload(dump: Union[CanDump, CanSend]):
    frame = dump.data
    size = len(frame)

    # global query_class

    if size < 2:
        return dali_payload.DaliResponse(last_queries.get(dump.channel), dump)
    elif size == 2:
        cmd = ext.from_frame(ext.ForwardFrame(16, frame), devicetype=device_type.get(dump.channel, 0))

        if isinstance(dump, DaliDump):
            if isinstance(cmd, ext.dali.gear.EnableDeviceType):
                device_type[dump.channel] = cmd.param
            else:
                device_type[dump.channel] = 0

        if cmd.is_query and isinstance(dump, DaliDump):
            last_queries[dump.channel] = cmd

        return dali_payload.DaliCommand(dump, cmd)
    elif size == 3:
        if frame[0] & 1 == 0:
            return dali_payload.Dali2Event(dump)
        else:
            cmd = ext.from_frame(ext.ForwardFrame(24, frame))
            if cmd.is_query and isinstance(dump, DaliDump):
                last_queries[dump.channel] = cmd
            return dali_payload.Dali2Command(dump, cmd)

    return Payload(dump)
