import time

from dali.address import Broadcast, Short
from dali.device import general as device
from dali.exceptions import NotConnected, BadDevice, NoFreeAddress

from resh.utils.errors import DeviceNotAnswered, CanError, CanCollision

MAX_DEVICE = 64


class Bus0(object):
    """A DALI bus."""
    Initialise = None
    Randomise = None
    Terminate = None
    Withdraw = None
    Compare = None
    ProgramShortAddress = None
    QueryShortAddress = None
    SetSearchAddrH = None
    SetSearchAddrM = None
    SetSearchAddrL = None
    CheckCrossedAddress = None
    CheckSelfAddress = None

    def __init__(self, interface, cross_addressing, start_index, step_handler=None):
        self.cross_addressing = cross_addressing
        self.start_index = start_index
        self._interface = interface
        self._step_handler = step_handler
        self._h = 0x0
        self._m = 0x0
        self._l = 0x0

    def query_short_address(self):
        return self.interface.send(self.QueryShortAddress()).value.as_integer

    def start_quiescent_mode(self):
        self.interface.send(device.StartQuiescentMode(Broadcast()))

    def stop_quiescent_mode(self):
        self.interface.send(device.StopQuiescentMode(Broadcast()))
        self.interface.send(self.Terminate())

    def set_search_addr(self, addr):
        i = self.interface

        h = (addr >> 16) & 0xff
        if self._h != h:
            self._h = h
            i.send(self.SetSearchAddrH(h))

        m = (addr >> 8) & 0xff
        if self._m != m:
            self._m = m
            i.send(self.SetSearchAddrM(m))

        l = addr & 0xff
        if self._l != l:
            self._l = l
            i.send(self.SetSearchAddrL(l))

    def assign_short_addresses(self, fill_addresses, **kwargs):
        """Search for devices on the bus with no short address allocated, and
        allocate each one a short address from the set of unused
        addresses.
        """
        new_devices = []
        self.start_quiescent_mode()
        free = self.unused_addresses(fill_addresses)
        i = self.interface
        i.send(self.Terminate())
        i.send(self.Initialise(*kwargs.values()))
        i.send(self.Randomise())
        time.sleep(0.1)
        low = 0
        high = 0xffffff
        while low is not None:
            low = self.find_next(low, high)
            if low is not None:
                if free:
                    new_addr = free.pop(0)
                    i.send(self.ProgramShortAddress(new_addr))
                    desc = None
                    try:
                        i.send(self.QueryShortAddress())
                    except DeviceNotAnswered:
                        desc = 'Устройство не ответило на запрос короткого адреса'
                    except (CanError, CanCollision):
                        desc = 'На запрос короткого адреса устройство ответило ошибкой'
                    except BaseException as ex:
                        desc = str(ex)
                    finally:
                        self.step_handler(action='found', value=new_addr, desc=desc)
                        self.step_handler(action='progress', value=min(99, int(len(new_devices) / 64 * 100)))
                        i.send(self.Withdraw())
                        new_devices.append(new_addr)
                else:
                    self.stop_quiescent_mode()
                    raise NoFreeAddress()
                low = low + 1
        self.stop_quiescent_mode()
        return new_devices

    def step_handler(self, action, value, desc=None):
        if self._step_handler:
            self._step_handler(action, value, desc)

    @property
    def interface(self):
        if not self._interface:
            raise NotConnected()
        return self._interface

    def unused_addresses(self, fill_addresses):
        """Return all short addresses that are not in use."""
        result = list(range(self.start_index, MAX_DEVICE))
        if fill_addresses:
            result = self.fill_free_addresses(result)

        if not self.cross_addressing:
            result = self.fill_free_addresses(result, crossed=True)
        return result

    def address_resolution(self, **initialize_params):
        self.start_quiescent_mode()
        devices = self.find_repetitive_short_addresses(**initialize_params)
        res = {}

        free = list(set(self.unused_addresses(True)) - devices.keys())

        for short, _arr in devices.items():
            if len(_arr) > 1:
                res[short] = []
                for long in _arr[1:]:
                    if free:
                        sa = free.pop(0)
                        self.set_search_addr(long)
                        self.interface.send(self.ProgramShortAddress(sa))
                        self.step_handler('fixed', [short, sa])
                        res[short].append(sa)
                    else:
                        self.stop_quiescent_mode()
                        raise NoFreeAddress()

        self.stop_quiescent_mode()
        return res

    def find_repetitive_short_addresses(self, **kwargs):
        """Search for devices on the bus with a repetitive short address."""

        devices = dict()
        i = self.interface
        i.send(self.Terminate())
        i.send(self.Initialise(*kwargs.values()))
        i.send(self.Randomise())
        time.sleep(0.1)
        low = 0
        high = 0xffffff
        while low is not None:
            low = self.find_next(low, high)
            if low is not None:
                short = self.query_short_address()
                i.send(self.Withdraw())
                if short == 127:
                    continue
                elif short in range(0, 64):
                    devices.setdefault(short, [])
                    devices[short].append(low)
                    if len(devices[short]) > 1:
                        self.step_handler('prefixed', short)
                    self.step_handler('progress', min(99, int(len(devices) / 64 * 100)))
                else:
                    self.stop_quiescent_mode()
                    raise BadDevice("device address is %s" % short)
                low = low + 1
        return devices

    def compare(self):
        try:
            return self.interface.send(self.Compare()).value
        except DeviceNotAnswered:
            return False
        except (CanError, CanCollision):
            return True

    def find_next(self, low, high):
        """Find the ballast with the lowest random address.  The caller
        guarantees that there are no ballasts with an address lower
        than 'low'.

        If found, returns the random address.  SearchAddr will be set
        to this address in all ballasts.  The ballast is not
        withdrawn.

        If not found, returns None.
        """
        self.set_search_addr(high)
        if low == high:
            if self.compare():
                return low
            return None
        # time.sleep(0.1)

        if self.compare() or self.compare():
            midpoint = (low + high) // 2
            return self.find_next(low, midpoint) or self.find_next(midpoint + 1, high)

    def fill_free_addresses(self, addresses, crossed=False):
        result = []
        for addr in addresses:
            try:
                if crossed:
                    self.interface.send(self.CheckCrossedAddress(Short(addr)))
                else:
                    self.interface.send(self.CheckSelfAddress(Short(addr)))
            except DeviceNotAnswered:
                result.append(addr)
            except (CanError, CanCollision):
                continue
        return result
