from resh.locales.loc_def import loc

NONEXISTENT_LOCALE = 1000
UNAVAILABLE_DOMAIN = 1001
UNAVAILABLE_METHOD = 1002
BAD_MODULE_ADDRESS = 1003
BAD_GATEWAY_INDEX = 1004
BAD_CHANNEL_ADDRESS = 1005
BAD_DEVICE_ADDRESS = 1006
BAD_INSTANCE_ADDRESS = 1007
BAD_PARAMETER = 1008
BAD_PARAMETER_ERROR = 1009
CAN_OPEN_ERROR = 1010
CAN_COMMIT_TIMEOUT = 1011
DEVICE_NOT_ANSWER = 1012
CAN_ERROR = 1013
CUSTOM_ERROR = 1014
UNAVAILABLE_INSTANCE = 1015
MODULE_CONSTRUCT_ERROR = 1016
ADDRESS_IS_BUSY = 1017
PACKET_PARSE_ERROR = 1018
NEGATIVE_RESULT = 1019


class ErrCode:
    can_open_error = dict(code=CAN_OPEN_ERROR, text='Can interface can not be opened', parameters=('reason',))
    can_commit_timeout = dict(code=CAN_COMMIT_TIMEOUT, text='Command not committed')
    can_device_not_answer = dict(code=DEVICE_NOT_ANSWER, text='Device not answered!')
    can_error = dict(code=CAN_ERROR, text='Can error')

    module_address = dict(code=BAD_MODULE_ADDRESS, text=loc.BAD_MODULE_ADDRESS)
    gateway_index = dict(code=BAD_GATEWAY_INDEX, text=loc.BAD_GATEWAY_INDEX)
    channel_address = dict(code=BAD_CHANNEL_ADDRESS, text=loc.BAD_CHANNEL)
    device_address = dict(code=BAD_DEVICE_ADDRESS, text=loc.BAD_DEVICE_ADDRESS)
    instance_address = dict(code=BAD_INSTANCE_ADDRESS, text=loc.BAD_INSTANCE)

    address_is_busy = dict(code=ADDRESS_IS_BUSY, text='Address({}) already in use. Try "replace_short_address"')


class DomainError(BaseException):
    show_help = True

    def __init__(self, err_type, *args, **kwargs):
        self.code = err_type['code']
        self.text = err_type['text']
        if 'stack' in kwargs:
            self._stack = kwargs.pop('stack')
        self.args = args
        self.parameters = kwargs

    @property
    def serialize(self):
        res = dict(error=dict(
            code=self.code,
            text=self.text
        ))

        if self.parameters:
            params = dict()
            for key, value in self.parameters.items():
                params[key] = value
            res['error']['parameters'] = params

        return res

    def __str__(self):
        text = self.text
        if self.args:
            text = text.format(*self.args)
        text = 'Error({0}): {1}'.format(self.code, text)

        if self.parameters:
            for key, value in self.parameters.items():
                text += '\n     {}: {}'.format(key, value)
        return text


class UnavailableDomain(DomainError):
    def __init__(self, domain):
        super(UnavailableDomain, self).__init__(dict(code=UNAVAILABLE_DOMAIN,
                                                     text=loc.UNAVAILABLE_DOMAIN.format(domain)))


class UnavailableMethod(DomainError):
    def __init__(self, method_name):
        super(UnavailableMethod, self).__init__(dict(code=UNAVAILABLE_METHOD,
                                                     text=loc.UNAVAILABLE_METHOD.format(method_name)))


class ModuleAddressError(DomainError):
    def __init__(self):
        super(ModuleAddressError, self).__init__(ErrCode.module_address)


class BadGatewayIndex(DomainError):
    def __init__(self):
        super(BadGatewayIndex, self).__init__(ErrCode.gateway_index)


class ChannelAddressError(DomainError):
    def __init__(self): super().__init__(ErrCode.channel_address)


class DeviceAddressError(DomainError):
    def __init__(self): super().__init__(ErrCode.device_address)


class InstanceAddressError(DomainError):
    def __init__(self): super().__init__(ErrCode.instance_address)


class NonexistentLocale(DomainError):
    def __init__(self, *keys):
        super(NonexistentLocale, self).__init__(dict(code=NONEXISTENT_LOCALE,
                                                     text=f'{loc.NONEXISTENT_LOC}' + ' > '.join([k for k in keys])))


class AddressAlreadyInUse(DomainError):
    def __init__(self, addr):
        super(AddressAlreadyInUse, self).__init__(dict(code=ADDRESS_IS_BUSY,
                                                       text=loc.ADDRESS_IS_BUSY.format(addr)))


class ParameterError(DomainError):
    def __init__(self, ex: BaseException):
        super(ParameterError, self).__init__(dict(code=BAD_PARAMETER_ERROR, text=str(ex)))


class BadParameter(DomainError):
    def __init__(self, message: str = loc.PARAMETER_ERROR):
        super(BadParameter, self).__init__(dict(code=BAD_PARAMETER_ERROR, text=message))


class UnavailableInstance(DomainError):
    def __init__(self):
        super(UnavailableInstance, self).__init__(dict(code=UNAVAILABLE_INSTANCE, text=loc.UNAVAILABLE_INSTANCE))


class CustomError(DomainError):
    def __init__(self, message):
        super(CustomError, self).__init__(dict(code=CUSTOM_ERROR, text=str(message)))


class NegativeResult(DomainError):
    show_help = False

    def __init__(self, message):
        super(NegativeResult, self).__init__(dict(code=NEGATIVE_RESULT, text=str(message)))


class ModuleConstructError(DomainError):
    show_help = False

    def __init__(self, message):
        super(ModuleConstructError, self).__init__(dict(code=MODULE_CONSTRUCT_ERROR, text=str(message)))


class DeviceNotAnswered(DomainError):
    show_help = False

    def __init__(self, **kwargs): super(DeviceNotAnswered, self).__init__(ErrCode.can_device_not_answer, **kwargs)

    @property
    def stack(self): return self._stack


class CanError(DomainError):
    show_help = False

    def __init__(self, **kwargs): super(CanError, self).__init__(ErrCode.can_error, **kwargs)

    @property
    def stack(self): return self._stack


class CanCollision(DomainError):
    show_help = False

    def __init__(self, **kwargs): super(CanCollision, self).__init__(ErrCode.can_error, **kwargs)

    @property
    def stack(self): return self._stack


class CanNotCommitted(DomainError):
    def __init__(self, **kwargs):
        super(CanNotCommitted, self).__init__(ErrCode.can_commit_timeout, **kwargs)


class CanOpenError(DomainError):
    def __init__(self, **kwargs):
        super(CanOpenError, self).__init__(ErrCode.can_open_error, **kwargs)


class PacketParseError(DomainError):
    def __init__(self, frame_str):
        super(PacketParseError, self).__init__(dict(code=PACKET_PARSE_ERROR, text=frame_str))


class StopExec(BaseException):
    pass
