import json
import signal
import sys
from datetime import datetime

from dali.command import BitmapResponse

from resh.utils.tools import StrByte

INFO = 1
CAN = 2


class Log:
    def __init__(self):
        self.log_level = 0
        self.is_json = False

    def set_log_level(self, level: int):
        self.log_level = level

    def set_json_out(self, is_json: bool):
        self.is_json = is_json

    @staticmethod
    def _print_with_time(message):
        print('{}: {}'.format(datetime.now(), message))

    def _print_data(self, key, data, tab_count=0):
        if isinstance(data, dict):
            print(f'{"    "*tab_count}{key}: ')
            for key, value in data.items():
                self._print_data(key, value, tab_count+1)
        elif isinstance(data, (list, tuple)):
            if key:
                print(f'{"    "*tab_count}{key}:')
                tab_count += 1
            for i in data:
                print(f'{"    " * tab_count}{i}')
        else:
            print(f'{"    "*tab_count}{key}: {data}')

    def _easy_print(self, root_name: str, data):
        if self.is_json:
            print(json.dumps({root_name: data}, default=str))
        else:
            if isinstance(data, dict):
                for key, value in data.items():
                    self._print_data(key, value)
            elif isinstance(data, (list, tuple)) and len(str(data)) > 70:
                self._print_data('', data)
            elif data is not None:
                print(f'{data}')

    def result(self, data):
        if isinstance(data, BitmapResponse):
            d = dict()
            v = data._value[7:0]
            for b in data.bits:
                b_name = b.strip()
                while ' ' in b_name:
                    space = b_name.index(' ')
                    if len(b_name) - 1 > space:
                        b_name = b_name[:space] + b_name[space + 1].upper() + b_name[space + 2:]
                    else:
                        break

                d[b_name] = bool(v & 0x01 and b)
                v = (v >> 1)
            data = d
        elif isinstance(data, (bytes, bytearray)):
            if len(data) == 1:
                data = StrByte(data)
            else:
                data = ' '.join(hex(b)[2:].rjust(2, '0') for b in data)

        self._easy_print('result', data)

    def monitor(self, message):
        if self.is_json:
            message = message.to_dict()
        self._easy_print('can_message', message)

    def can(self, message):
        if CAN == self.log_level:
            self._print_with_time(message)

    def info(self, message):
        if INFO == self.log_level:
            self._print_with_time(message)

    def exception(self, ex: BaseException):
        if self.is_json:
            if hasattr(ex, 'serialize'):
                print(json.dumps(ex.serialize))
            else:
                print(json.dumps(str(ex)))
        else:
            print(str(ex).strip())

    def question(self, message, timeout=999):
        def alarm_handler(signum, frame):
            raise TimeoutError()
        signal.signal(signal.SIGALRM, alarm_handler)
        signal.alarm(timeout)
        try:
            return input(f'{message}>').strip()
        except:
            return None
        finally:
            sys.stdout.write('\r' + ' '*(len(message) + 1) + '\r')
            signal.alarm(0)


log = Log()
