from dali import address
from dali import command
from dali.device import general as device
from dali.frame import BackwardFrame
from dali.gear import general as gear

from resh.protocols.payloads import Payload
from resh.utils import tools


class DaliPayload(Payload):
    @property
    def has_address(self):
        return False

    @property
    def has_instance(self):
        return False

    @property
    def address(self):
        raise ValueError()

    @property
    def instance(self):
        raise ValueError()

    @property
    def destination_str(self):
        arr = [super(DaliPayload, self).destination_str, self.dump.channel]
        if self.has_address:
            arr.append(self.address.ljust(3, ' '))
        if self.has_instance:
            arr.append(self.instance)
        return ' '.join(str(i) for i in arr)


class DaliCommand(DaliPayload):
    mark = 'dali'

    @property
    def has_address(self):
        return isinstance(self.command, (gear._StandardCommand, gear.DAPC))

    @property
    def has_params(self):
        return isinstance(self.command, gear.DAPC) \
               or (isinstance(self.command, (gear._SpecialCommand, gear._StandardCommand)) and self.command._hasparam)

    @property
    def params(self):
        if isinstance(self.command, gear.DAPC):
            return [self.command.power]
        return [self.command.param]

    @property
    def has_instance(self):
        return False

    @property
    def instance(self):
        return None

    @property
    def address(self):
        return tools.pack_address(self.command.destination)

    @property
    def name(self) -> str:
        arr = [self.command.__class__.__name__]
        if self.has_params:
            arr.append(' '.join([str(p) for p in self.params]))
        return ' '.join(str(i) for i in arr)


class Dali2Command(DaliCommand):
    mark = 'dali2'

    @property
    def has_address(self):
        return isinstance(self.command, device._DeviceCommand) and not isinstance(self.command, device._SpecialDeviceCommand)

    @property
    def has_instance(self):
        return isinstance(self.command, device._StandardInstanceCommand)

    @property
    def has_params(self):
        return isinstance(self.command, (device._SpecialDeviceCommandOneParam, device._SpecialDeviceCommandTwoParam))

    @property
    def params(self):
        if isinstance(self.command, device._SpecialDeviceCommandOneParam):
            return [self.command._opcode]
        elif isinstance(self.command, device._SpecialDeviceCommandTwoParam):
            return [self.command._instance, self.command._opcode]
        else:
            return []

    @property
    def instance(self):
        return tools.pack_instance(self.command.instance)


class DaliResponse(DaliPayload):
    def __init__(self, query: command.Command, *args, **kwargs):
        super(DaliResponse, self).__init__(*args, **kwargs)
        self.query = query

    @property
    def is_dali2(self):
        return issubclass(self.query.__class__, device._DeviceCommand)

    @property
    def has_address(self):
        if isinstance(self.query, gear._GearCommand):
            return isinstance(self.query, (gear._StandardCommand, gear.DAPC))
        elif isinstance(self.query, device._DeviceCommand):
            return not isinstance(self.query, device._SpecialDeviceCommand)
        return False

    @property
    def address(self):
        return tools.pack_address(self.query.destination)

    @property
    def instance(self):
        return tools.pack_instance(self.query.instance)

    @property
    def has_instance(self):
        return isinstance(self.query, device._StandardInstanceCommand)

    @property
    def name(self) -> str:
        if self.dump.is_error:
            result = 'error'
        elif len(self.dump.data) == 0:
            result = 'empty'
        elif self.query and self.query.response:
            frame = BackwardFrame(self.dump.data[0])
            resp = self.query._response(frame)
            if isinstance(resp, (gear.QueryFadeTimeAndRateResponse, gear.QueryDeviceTypeResponse)):
                result = str(resp)
            elif isinstance(resp, command.BitmapResponse):
                return f'{resp.__class__.__name__} {resp.status}'
            elif isinstance(resp, command.YesNoResponse):
                result = resp.value
            elif isinstance(self.query, gear.QueryShortAddress):
                result = resp.value.as_integer >> 1
            else:
                result = resp.value.as_integer
        else:
            return str(self.dump)

        return f'{self.__class__.__name__} {result}'

    @property
    def cmd_name(self) -> str:
        if self.query and self.query.response:
            return self.query.__class__.__name__

        return self.__class__.__name__

    @property
    def mark(self):
        if self.query and self.query.response:
            if isinstance(self.query, device._DeviceCommand):
                return Dali2Command.mark
            elif isinstance(self.query, gear._GearCommand):
                return DaliCommand.mark

        return DaliPayload.mark


class Dali2Event(DaliPayload):
    mark = 'event'

    BUTTON_RELEASED = 'ButtonReleased'
    BUTTON_PRESSED = 'ButtonPressed'
    SHORT_PRESS = 'ShortPress'
    DOUBLE_PRESS = 'DoublePress'
    LONG_PRESS_START = 'LongPressStart'
    LONG_PRESS_REPEAT = 'LongPressRepeat'
    LONG_PRESS_STOP = 'LongPressStop'
    BUTTON_FREE = 'ButtonFree'
    BUTTON_STUCK = 'ButtonStuck'
    DEFAULT = 'UnknownButtonEvent'

    STATES = {
        0: BUTTON_RELEASED,
        1: BUTTON_PRESSED,
        2: SHORT_PRESS,
        5: DOUBLE_PRESS,
        9: LONG_PRESS_START,
        11: LONG_PRESS_REPEAT,
        12: LONG_PRESS_STOP,
        14: BUTTON_FREE,
        15: BUTTON_STUCK,
    }

    d1 = d2 = val = _address = _instance = None

    def __init__(self, *args, **kwargs):
        super(Dali2Event, self).__init__(*args, **kwargs)
        self.__parse()

    @property
    def has_address(self):
        return self._address is not None

    @property
    def address(self):
        return tools.pack_address(self._address)

    @property
    def has_instance(self):
        return self._instance is not None

    @property
    def instance(self):
        return tools.pack_instance(self._instance)

    @property
    def destination_str(self):
        return ' '.join(str(i) for i in [self.dump.module_address, self.dump.channel, self.d1.ljust(3, ' '), self.d2])

    @property
    def name(self) -> str:
        return f'{self.__class__.__name__} {self.val}'

    def __parse(self):
        frame = self.dump.data

        if frame[0] >> 7 == 0x0:
            short = frame[0] >> 1
            self.d1 = f'D{short}'
            self._address = address.Short(short)
            if frame[1] >> 7 == 0x0:
                tp = frame[1] >> 2
                self._instance = address.InstanceType(tp)
                self.d2 = f'IT{tp}'
            elif frame[1] >> 7 == 0x1:
                number = (frame[1] & 0x7f) >> 2
                self._instance = address.InstanceNumber(number)
                self.d2 = f'IN{number}'
        elif frame[0] >> 6 == 0x2:
            if frame[1] >> 7 == 0x0:
                group = (frame[0] & 0x3f) >> 1
                self.d1 = f'G{group}'
                self._address = address.Group(group)
                tp = (frame[1] & 0x7f) >> 2
                self.d2 = f'IT{tp}'
                self._instance = address.InstanceType(tp)
            elif frame[1] >> 7 == 0x1:
                self.tp = (frame[0] & 0x3f) >> 1
                self._instance = address.InstanceType(self.tp)
                self.d1 = f'IT{self.tp}'
                self.d2 = f'IN{(frame[1] & 0x7f) >> 2}'
        elif frame[0] >> 6 == 0x3:
            if frame[1] >> 7 == 0x0:
                self.d1 = f'IG{(frame[0] & 0x3f) >> 1}'
                self.tp = (frame[1] & 0x7f) >> 2
                self.d1 = f'IT{self.tp}'
                self._instance = address.InstanceType(self.tp)

        if isinstance(self._instance, address.InstanceType) and self._instance._value == 3:
            self.val = self.dump.data[2] & 0x03 != 0x00
        elif isinstance(self._instance, address.InstanceType) and self._instance._value == 1:
            self.val = self.STATES.get(self.dump.data[2], self.DEFAULT)
        else:
            self.val = int.from_bytes(self.dump.data, 'big') & int('1'*10, 2)
