from datetime import datetime

from resh.protocols.can_utils import CONTROLLER, POWER
from resh.protocols.can_utils.can_packet import CanPacket
from resh.utils.tools import bytes2str

HEADER_OFFSET = 20
DIRECTION_OFFSET = 12
DESTINATION_OFFSET = 11


class Payload:
    mark = 'Payload'

    def __init__(self, dump: CanPacket, command=None):
        self.dump = dump
        self.command = command

    def __repr__(self):
        return f'[{self.destination_str} {self.name}]: {bytes2str(self.dump.frame)}'

    def __str__(self):
        return self.header_str() + ' ' + ' '.join(str(p) for p in [
            self.direction.ljust(DIRECTION_OFFSET, ' '),
            self.destination_str.ljust(DESTINATION_OFFSET, ' '),
            ' ',
            self.name
        ])

    def header_str(self) -> str:
        return ' '.join(str(p) for p in [
            self.time + ':',
            self.raw_hex.ljust(HEADER_OFFSET, ' ')
        ])

    def to_dict(self):
        return dict(
            time=self.time,
            hex=bytes2str(self.dump.data, " "),
            from_id=self.dump.from_id,
            to_id=self.dump.to_id,
            cmd=self.dump.cmd,
            module_type=self.dump.module_type,
            module=self.dump.module_address,
            name=self.name
        )

    @property
    def time(self):
        time = datetime.now()
        return time.strftime('%H:%M:%S') + f'.{str(int(time.microsecond / 1000)).ljust(3, "0")}'

    @property
    def destination_str(self):
        return f'{self.dump.module_address}'

    @property
    def name(self) -> str:
        if self.command is not None:
            if isinstance(self.command, type):
                return self.command.__name__
            else:
                return self.command.__class__.__name__

        return self.__class__.__name__

    @property
    def cmd_name(self) -> str:
        if self.command is not None:
            if isinstance(self.command, type):
                return self.command.__name__
            else:
                return self.command.__class__.__name__

        return self.__class__.__name__

    @property
    def direction(self):
        _from = self.dump.from_id
        _to = self.dump.to_id
        _a1 = _a2 = ''
        _dir = '->'

        if _from == CONTROLLER:
            _a1 = 'C'
            _a2 = ('*' if self.dump.cmd == 1 else '') + self.dump.module_name
        elif _to == CONTROLLER:
            _a1 = 'C'
            _a2 = ('*' if self.dump.cmd == 1 else '') + self.dump.module_name
            _dir = '<-'
        elif _from == POWER:
            _a1 = 'P'
            _a2 = ('*' if self.dump.cmd == 1 else '') + self.dump.module_name
        elif _to == POWER:
            _a1 = 'P'
            _a2 = ('*' if self.dump.cmd == 1 else '') + self.dump.module_name
            _dir = '<-'
        else:
            _a1 = ('*' if self.dump.cmd == 1 else '') + str(_from)
            _a2 = _to

        return f'{str(_a1).rjust(2, " ")} {_dir} {str(_a2).ljust(2, " ")}'

    @property
    def raw_hex(self):
        return f'{hex(self.dump.arbitration_id)[2:].rjust(3, "0")}#{bytes2str(self.dump.frame, "")}'.upper()
